from typing import Union

import numpy as np
import ramanchada2 as rc2
from AnyQt.QtWidgets import QGroupBox
from Orange.widgets import gui
from pydantic import validate_call


class GenSpe:
    @validate_call(config=dict(arbitrary_types_allowed=True))
    def __init__(self, parent, *,
                 spe_xmin: tuple[str, Union[None, QGroupBox]],
                 spe_xmax: tuple[str, Union[None, QGroupBox]],
                 spe_nbins: tuple[str, Union[None, QGroupBox]],
                 deltas: tuple[str, Union[None, QGroupBox]],
                 ):
        self._parent = parent

        self._spe_xmin = spe_xmin[0]
        if spe_xmin[1]:
            gui.spin(spe_xmin[1], self._parent, self._spe_xmin, -1000, 10000, label='xmin', callback=self.auto_process)

        self._spe_xmax = spe_xmax[0]
        if spe_xmax[1]:
            gui.spin(spe_xmax[1], self._parent, self._spe_xmax, -1000, 10000, label='xmax', callback=self.auto_process)

        self._spe_nbins = spe_nbins[0]
        if spe_nbins[1]:
            gui.spin(spe_nbins[1], self._parent, self._spe_nbins, 1, 200000, label='n_bins', callback=self.auto_process)

        self._deltas = deltas[0]
        self._deltas_combo = deltas[0] + '_combo'
        if deltas[1]:
            self.deltas_edit = gui.lineEdit(deltas[1], self._parent, self._deltas, label='Deltas',
                                            callback=self.auto_process)
            gui.comboBox(deltas[1], self._parent, self._deltas_combo, sendSelectedValue=True,
                         items=['PST', 'User defined'], callback=self.set_deltas)
            self.set_deltas()

    def set_deltas(self):
        if self.deltas_combo == 'User defined':
            self.deltas_edit.setReadOnly(False)
        else:
            self.deltas = '620.9: 16, 795.8: 10, 1001.4: 100, 1031.8: 27, 1155.3: 13, 1450.5: 8, 1583.1: 12, 1602.3: 28, 2852.4: 9, 2904.5: 13, 3054.3: 32'  # noqa: E501
            self.deltas_edit.setReadOnly(True)

    @validate_call(config=dict(arbitrary_types_allowed=True))
    def __call__(self) -> rc2.spectrum.Spectrum:
        deltas_dict = dict([[float(j) for j in i.split(':')] for i in self.deltas.replace(' ', '').split(',')])
        spe1 = rc2.spectrum.Spectrum(x=np.array(list(deltas_dict.keys())), y=np.array(list(deltas_dict.values())))
        return spe1.resample_NUDFT_filter(x_range=(self.spe_xmin, self.spe_xmax), xnew_bins=self.spe_nbins)

    def auto_process(self):
        self._parent.auto_process()

    @property
    def spe_xmin(self):
        return getattr(self._parent, self._spe_xmin)

    @spe_xmin.setter
    def spe_xmin(self, val):
        setattr(self._parent, self._spe_xmin, val)

    @property
    def spe_xmax(self):
        return getattr(self._parent, self._spe_xmax)

    @spe_xmax.setter
    def spe_xmax(self, val):
        setattr(self._parent, self._spe_xmax, val)

    @property
    def spe_nbins(self):
        return getattr(self._parent, self._spe_nbins)

    @spe_nbins.setter
    def spe_nbins(self, val):
        setattr(self._parent, self._spe_nbins, val)

    @property
    def deltas(self):
        return getattr(self._parent, self._deltas)

    @deltas.setter
    def deltas(self, val):
        setattr(self._parent, self._deltas, val)

    @property
    def deltas_combo(self):
        return getattr(self._parent, self._deltas_combo)

    @deltas_combo.setter
    def deltas_combo(self, val):
        setattr(self._parent, self._deltas_combo, val)




import yaml
import os.path
import pandas
import tempfile
import pandas as pd
from Orange.data.pandas_compat import table_from_frame

def load_env(env_file,name="ploomber_twinning"):
    with open(env_file, "r") as file:
        env = yaml.safe_load(file)        
    env["output_folder"] = os.path.join(tempfile.gettempdir(),name)
    return env

def env2table(env):
    return table_from_frame(pd.DataFrame.from_dict(env, orient="index", columns=["value"]))


from Orange.widgets import gui
from Orange.widgets.settings import Setting
from Orange.widgets.widget import Msg
from scipy import signal
import numpy as np
from ramanchada2.spectrum import Spectrum
import pandas as pd
from Orange.data.pandas_compat import table_from_frame
from Orange.data import Domain, ContinuousVariable, Table,StringVariable,DiscreteVariable


from ..base_widget import FilterWidget


class Resample_density(FilterWidget):
    name = "Resample"
    description = "Resample using density"
    icon = "icons/spectra.svg"

    xmin = Setting(100)
    xmax = Setting(3600)
    npoints = Setting(3500)
    show_original = Setting(False)

    class Warning(FilterWidget.Warning):
        range = Msg('x-min should be less that x-max')

    def __init__(self):
        super().__init__()
        box1 = gui.widgetBox(self.controlArea, "Resampled x axis")
        gui.spin(box1, self, 'xmin', -1000, 10000, callback=self.auto_process, label='x-min')
        gui.spin(box1, self, 'xmax', -1000, 10000, callback=self.auto_process, label='x-max')
        gui.spin(box1, self, 'npoints', 1, 10000, callback=self.auto_process, label='npoints')
        box2 = gui.widgetBox(self.controlArea, "Visualisaiton")
        gui.checkBox(box2, self, "show_original", "Show original", callback=self.auto_process)

    def process(self):
        self.Warning.clear()
        self.out_spe = list()
        x_values = np.linspace(self.xmin, self.xmax, self.npoints)
        for spe in self.in_spe:
            if self.xmin > self.xmax:
                self.Warning.range()
            else:
                dist = spe.spe_distribution(trim_range=(self.xmin, self.xmax))

                #resample using probability density function
                y_values = dist.pdf(x_values)
                scale = np.max(spe.y) / np.max(y_values)
                 # pdf sampling is normalized to area unity, scaling back
                _tmp = y_values *  scale
                self.out_spe.append(
                    Spectrum(x_values,_tmp,metadata=spe.meta)
                    )
        self.send_outputs()


    def custom_plot(self, ax):
        if self.show_original:
            for spe in self.in_spe:
                spe.trim_axes(method="x-axis",boundaries=(self.xmin, self.xmax)).plot(ax=ax,label='original') 
        ax.legend()

    def old_send_output_table(self):
        if self.should_pass_datatable:
            # Extract unique x values from all spectra
            unique_x_values = sorted(set(x for spe in self.out_spe for x, _ in zip(spe.x, spe.y)))
            file_ids = []
            for spe in self.out_spe:
                try:
                    meta_info = spe.meta["Original file"]
                except  KeyError:
                    meta_info = str(id(spe))
                file_ids.append(meta_info)

              # Create Orange Table
            domain = Domain([ContinuousVariable(str(x)) for x in unique_x_values] 
                            # + [DiscreteVariable("ID", values=set(file_ids))]
                            , metas=[StringVariable("ID")]
                            #,class_vars=[DiscreteVariable("ID", values=set(file_ids))]
                            )
            data = []

            # Iterate through each spectrum to populate the data
            for spe, file_id in zip(self.out_spe, file_ids):
                row_data = [0] * len(unique_x_values)  # Initialize row data with zeros
                for x, y in zip(spe.x, spe.y):
                    column_index = unique_x_values.index(x)
                    row_data[column_index] = y
                data.append(row_data + [file_id])

            table = Table(domain, data)
            self.Outputs.data.send(table)


import pickle
from Orange.widgets import widget, gui
from Orange.widgets.widget import Input, Output
from ramanchada2.protocols.calibration import CalibrationModel
from AnyQt.QtWidgets import QFileDialog

class LoadCalibrationModel(widget.OWWidget):
    name = "x-calibration Load"
    description = "Load x-Calibration Model"
    icon = "icons/spectra.svg"
    priority = 10

    class Outputs:
        calmodel = Output("Calibration Model", CalibrationModel)

    def __init__(self):
        super().__init__()
        self.calmodel = None

        # Create load button
        self.load_button = gui.button(self.controlArea, self, "Load", callback=self.load_file)


    def load_file(self):
        options = QFileDialog.Options()
        filename, _ = QFileDialog.getOpenFileName(self, "Load x-calibration model", "",
                                                  "Calibration model (*.calmodel)", options=options)

        if filename:
            with open(filename, 'rb') as file:
                self.calmodel = pickle.load(file)
                self.Outputs.calmodel.send(self.calmodel)               

    def custom_plot(self, ax):
        self.calibration_model.plot(ax=self.ax)            



from Orange.widgets.widget import Input

from .base_widget import BaseWidget
from .types import RC2Spectra


class FilterWidget(BaseWidget, openclass=True):

    def input_hook(self):
        pass

    class Inputs:
        in_spe = Input("RC2Spectra", RC2Spectra, default=True, auto_summary = False)

    @Inputs.in_spe
    def set_in_spe(self, spe):
        self.Warning.clear()
        self.should_auto_plot = False
        if spe:
            self.in_spe = spe
            self.auto_process()
            self.info.set_input_summary(f'{len(self.in_spe)} RC2Spectra',
                                        '\n'.join([f'· {repr(i)}' for i in self.in_spe]))
        else:
            self.in_spe = RC2Spectra()
            self.info.set_input_summary(self.info.NoInput)
        self.input_hook()




from Orange.widgets import gui
from Orange.widgets.settings import Setting
from Orange.widgets.widget import  Input,Output, Msg, OWWidget
from ..base_widget import BaseWidget , FilterWidget, RC2Spectra
import ramanchada2 as rc2
from Orange.data import Table
import pickle
from AnyQt.QtWidgets import QFileDialog

from ramanchada2.protocols.calibration import CalibrationModel
available_models = ['Gaussian', 'Lorentzian', 'Moffat', 'Voigt', 'PseudoVoigt', 'Pearson4', 'Pearson7']

class XAxisCalibrationWidget(FilterWidget):
    name = "CHARISMA X axis calibration"
    description = "X-axis calibration according to CHARISMA protocol"
    icon = "icons/spectra.svg"
    laser_wl = Setting(785)

    kw_findpeak_prominence = Setting(3)
    kw_findpeak_wlen = Setting(200)
    kw_findpeak_width = Setting(1)
    
    ne_peak_profile = Setting(available_models[0])
    si_peak_profile = Setting(available_models[5])

    should_auto_proc = Setting(False)
    should_auto_plot = Setting(False)

    ne_should_fit = Setting(False)
    si_should_fit = Setting(True)


    def input_hook(self):
        pass

    class Warning(OWWidget.Warning):
        warning = Msg("Warning")

    class Information(OWWidget.Information):
        success = Msg("Workflow successful")

    class Error(OWWidget.Error):
        processing_error = Msg("Workflow error")

    class Inputs(FilterWidget.Inputs):
        spe_neon = Input("Neon spectrum", RC2Spectra, default=True)
        spe_si = Input("Si spectrum", RC2Spectra, default=False)

    class Outputs(FilterWidget.Outputs):
        out_model = Output("Calibration model", CalibrationModel)

    @Inputs.in_spe
    def set_in_spe(self, spe):
        if spe:
            self.in_spe = spe
        else:
            self.in_spe = RC2Spectra()
        self.update_inputs_info()

    @Inputs.spe_neon
    def set_spe_neon(self, spe):
        if spe:
            self.spe_neon = spe
        else:
            self.spe_neon = RC2Spectra()
        self.update_inputs_info()

    @Inputs.spe_si
    def set_spe_si(self, spe):
        if spe:
            self.spe_si = spe
        else:
            self.spe_si = RC2Spectra()
        self.update_inputs_info()        

    def update_inputs_info(self):
        self.Warning.clear()
        self.should_auto_plot = False        
        len00 = len(self.in_spe) if self.in_spe else None
        len11 = len(self.spe_neon) if self.spe_neon else None
        len12 = len(self.spe_si) if self.spe_si else None
       
        self.info.set_input_summary(f' {len00} RC2Spectra + Neon {len11} RC2Spectra + {len12} Si RC2Spectra')        

    def __init__(self):
        super().__init__()
        self.spe_neon = RC2Spectra()        
        self.spe_si = RC2Spectra()

        self.calibration_model = None
        box = gui.widgetBox(self.controlArea, self.name)
        gui.doubleSpin(box, self, 'laser_wl', 0, 5000, decimals=0, step=1, 
                       #callback=self.auto_process,
                       label='Laser Wavelength [nm]')    
        curve_box = gui.widgetBox(self.controlArea, "Calibration curve")     
        gui.comboBox(curve_box, self, 'ne_peak_profile', sendSelectedValue=True, items=available_models,
                     callback=self.auto_process,label="Peak profile")        
        gui.checkBox(curve_box, self, "ne_should_fit", "Should fit", callback=self.auto_process)

        lazerzero_box = gui.widgetBox(self.controlArea, "Lazer zeroing")   
        gui.comboBox(lazerzero_box, self, 'si_peak_profile', sendSelectedValue=True, items=available_models,
                     callback=self.auto_process)
        gui.checkBox(lazerzero_box, self, "si_should_fit", "Should fit", callback=self.auto_process)

        self.peakbox = gui.widgetBox(self.controlArea, "Peak finding options")     
        gui.doubleSpin(self.peakbox, self, 'kw_findpeak_prominence', 1, 10, decimals=1, step=1, callback=self.auto_process,
                       label='prominence')
        gui.doubleSpin(self.peakbox, self, 'kw_findpeak_wlen', 1, 1000, decimals=0, step=10, callback=self.auto_process,
                       label='wlen') 
        gui.doubleSpin(self.peakbox, self, 'kw_findpeak_width', 1, 10, decimals=1, step=1, callback=self.auto_process,
                       label='width')            

        self.save_button = gui.button(self.controlArea, self, "Save calibration model", callback=self.save_to_pickle)

    def derive_model(self,laser_wl,spe_neon,spe_sil):
        calmodel = CalibrationModel(laser_wl)
        calmodel.prominence_coeff = self.kw_findpeak_prominence
        print("derive_model_curve")
        find_kw = {"prominence" :spe_neon.y_noise * calmodel.prominence_coeff , "wlen" : self.kw_findpeak_wlen, "width" :  self.kw_findpeak_width }
        model_neon = calmodel.derive_model_curve(spe_neon,calmodel.neon_wl[laser_wl],spe_units="cm-1",ref_units="nm",find_kw=find_kw,fit_peaks_kw={},should_fit = self.ne_should_fit,name="Neon calibration")
        spe_sil_ne_calib = model_neon.process(spe_sil,spe_units="cm-1",convert_back=False)
        find_kw = {"prominence" :spe_sil_ne_calib.y_noise * calmodel.prominence_coeff , "wlen" : self.kw_findpeak_wlen, "width" :  self.kw_findpeak_width }
        print("derive_model_zero")
        model_si = calmodel.derive_model_zero(spe_sil_ne_calib,ref={520.45:1},spe_units="nm",ref_units="cm-1",
                            find_kw=find_kw,fit_peaks_kw={},should_fit=self.si_should_fit,name="Si calibration",profile=self.si_peak_profile)
        #model_si.peaks.to_csv(os.path.join(config_root,template_file.replace(".xlsx","peaks.csv")),index=False)
        #spe_sil_calib = model_si.process(spe_sil_ne_calib,spe_units="nm",convert_back=False)
        return calmodel        
    
    def apply_calibration_x(self, old_spe: rc2.spectrum.Spectrum, spe_units="cm-1"):
        new_spe = old_spe
        model_units = spe_units
        for model in self.calibration_model.components:
            if model.enabled:
                new_spe = model.process(new_spe, model_units, convert_back=False)
                model_units = model.model_units
        return new_spe    
    
    def process(self):
        #these should be done in previous widgets
        #self.spe_si[0] = self.spe_si[0].trim_axes(method='x-axis',boundaries=(520.45-200,520.45+200))
        #self.spe_neon[0] = self.spe_neon[0].trim_axes(method='x-axis',boundaries=(100,max(self.spe_neon[0].x)))

        ## baseline  SNIP
        #kwargs = {"niter" : 40 }
        #self.spe_neon[0] = self.spe_neon[0].subtract_baseline_rc1_snip(**kwargs)  
        #self.spe_si[0] = self.spe_si[0].subtract_baseline_rc1_snip(**kwargs)          
        self.calibration_model = self.derive_model(self.laser_wl,self.spe_neon[0],self.spe_si[0])

        self.out_spe = list()
        for spe in self.in_spe:
            self.out_spe.append(
                    self.apply_calibration_x(spe)
                )
        self.is_processed = True            
        self.send_outputs()    
    
    def custom_plot(self, ax):
        self.calibration_model.plot(ax=self.axes[0])            
        self.axes[0].legend()
        for spe in self.spe_si:
            spe.plot(ax=self.axes[1]) 
            _tmp = self.apply_calibration_x(spe)
            _tmp.plot(ax=self.axes[1],color='orange', label='calibrated')
        self.axes[1].legend()
        ax.set_xlabel("cm-1")

        self.axes[1].set_xlim(520.45-50,520.45+50)        
        if self.in_spe:
            for spe in self.in_spe:
                spe.plot(ax=self.axes[2],label="original")
    

    def plot_create_axes(self):
        self.axes = self.figure.subplots(nrows=3, sharex=False)
        return self.axes[2]
    
    def save_to_pickle(self):
        if self.calibration_model is not None:
            options = QFileDialog.Options()
            filename, _ = QFileDialog.getSaveFileName(self, "Save Calibration model", "",
                                                      "Calibration model (*.calmodel)", options=options)
            if filename:
                with open(filename, 'wb') as file:
                    pickle.dump(self.calibration_model, file)    
        
if __name__ == "__main__":
    from Orange.widgets.utils.widgetpreview import WidgetPreview
    import os
    try:
        WidgetPreview(XAxisCalibrationWidget).run()
    except Exception as err:
        print(err)
        WidgetPreview(XAxisCalibrationWidget).run()



# + tags=["parameters"]
upstream = []
product = None
root_led_folder: None
root_data_folder: None
files_led_reference: None
files_led_twinned: None
files_spectra_reference: None
files_spectra_twinned: None
filter_probe: None
input_rcspectra: None
key_spectra: None
key_leds: None

# -

import pandas as pd
import ramanchada2 as rc2
import os.path

A =[]
for filename in os.listdir(os.path.join(root_data_folder,files_spectra_reference)):
    try:
        if filter_probe in filename:
            spea=rc2.spectrum.from_local_file(os.path.join(root_data_folder,files_spectra_reference,filename))
            A.append(spea)
    except:
        pass
B=[]
for filename in os.listdir(os.path.join(root_data_folder,files_spectra_twinned)):
    try:
        if filter_probe in filename:        
            speb=rc2.spectrum.from_local_file(os.path.join(root_data_folder,files_spectra_twinned,filename))
            B.append(speb)
    except:
        pass

dfA= pd.DataFrame(A,columns=["spectrum"])
dfA["reference"]= True
dfB= pd.DataFrame(B,columns=["spectrum"])
dfB["reference"]= False
df_spectra = pd.concat([dfA,dfB], ignore_index=True)

df_spectra.to_hdf(product["data"], key=key_spectra, mode='w')

A =[]
for filename in os.listdir(os.path.join(root_led_folder,files_led_reference)):
    try:
        if filter_probe in filename:
            spea=rc2.spectrum.from_local_file(os.path.join(root_led_folder,files_led_reference,filename))
            A.append(spea)
    except:
        pass
B=[]
for filename in os.listdir(os.path.join(root_led_folder,files_led_twinned)):
    try:
        if filter_probe in filename:
            speb=rc2.spectrum.from_local_file(os.path.join(root_led_folder,files_led_twinned,filename))
            B.append(speb)
    except:
        pass

dfA= pd.DataFrame(A,columns=["spectrum"])
dfA["reference"]= True
dfB= pd.DataFrame(B,columns=["spectrum"])
dfB["reference"]= False
df_leds = pd.concat([dfA,dfB], ignore_index=True)

df_leds.to_hdf(product["data"], key=key_leds, mode='a')


from Orange.widgets import gui
from Orange.widgets.settings import Setting

from ..base_widget import FilterWidget


class RecoverSpikes(FilterWidget):
    name = "Recover Spikes"
    description = "Recover single-bin spikes using linear interpolation"
    icon = "icons/spectra.svg"

    sigma = Setting(10)

    def __init__(self):
        super().__init__()
        box = gui.widgetBox(self.controlArea, self.name)
        gui.spin(box, self, 'sigma', 1, 100, callback=self.auto_process)

    def process(self):
        self.out_spe = list()
        for spe in self.in_spe:
            self.out_spe.append(
                spe.recover_spikes(self.sigma)
                )
        self.send_outputs()



# + tags=["parameters"]
upstream = None
product = None
root_folder = None
config_input = None
# -


import pandas as pd
import os.path
import json
import glob


def read_opticalpath(sheet_name,file_metadata):

    tag_sample=["S0N","S0B","S0P","S1N","nCAL","sCAL","PST","Sil"]
    op_sheet = pd.read_excel(file_metadata,sheet_name=sheet_name,header=None)
    start_row = 29
    offset_row = 33
    tmp = []
    try:
        while not pd.isna(op_sheet.iloc[start_row][0]):

            for sample in range(0,len(tag_sample),1):
                file_name = op_sheet.iloc[start_row+sample][11]
                if not pd.isna(file_name):
                    files = file_name.strip().replace("\n",",").split(",")
                    files = [s.strip() for s in files]
                    tmp.append({"sample" : tag_sample[sample],"file" : files , "laser_power" : op_sheet.iloc[start_row][4]})
            start_row = start_row + offset_row
            if start_row>op_sheet.shape[0]:
                break
    except Exception as err:
        print(err,sheet_name)
        pass

    return tmp

def xlsx_to_dict(folder_input, file_metadata):
    front_sheet = pd.read_excel(file_metadata,sheet_name="Front sheet",header=None)
    front_sheet.fillna("",inplace=True)

    metadata={  "folder_input" : folder_input,
                "file_metadata" : file_metadata,
                "provider" : front_sheet.iloc[0][1],
               "instrument" : front_sheet.iloc[0][6], "wavelength" : front_sheet.iloc[1][6],
               "optical_path" : [] }

    ops = metadata["optical_path"]

    row = 6
    #while not pd.isna(front_sheet.iloc[row][0]):
    try:
        while (front_sheet.iloc[row][0]!=""):
            op_id = front_sheet.iloc[row][0]
            op = {"id" : op_id,
                    "collection_optics" : front_sheet.iloc[row][2],
                    "slit_size" : front_sheet.iloc[row][4],
                    "gratings" : front_sheet.iloc[row][6],
                    "pin_hole_size" : front_sheet.iloc[row][8],
                    "collection_fibre_diameter" : front_sheet.iloc[row][10],
                    "notes" : front_sheet.iloc[row][12]}
            #print(op)
            op["files"] = read_opticalpath(op_id,file_metadata)
            #print(op)
            ops.append(op)
            row=row+1
    except Exception as err:
        pass

    return metadata


def get(folder_input,file_metadata,product_metadata):
    metadata = xlsx_to_dict(folder_input, file_metadata)
    with open(product_metadata, "w",encoding="utf-8") as write_file:
        json.dump(metadata, write_file, sort_keys=True, indent=4)

def metadata_flatten(data):
    return [
            dict(
                file_metadata=data['file_metadata'],
                instrument=data['instrument'],
                collection_optics=optical_path['collection_optics'],
                collection_fibre_diameter=optical_path['collection_fibre_diameter'],
                file=file,
                laser_power=files['laser_power'],
                sample=files['sample'],
                gratings=optical_path['gratings'],
                id=optical_path['id'],
                notes=optical_path['notes'],
                pin_hole_size=optical_path['pin_hole_size'],
                slit_size=optical_path['slit_size'],
                provider=data['provider'],
                wavelength=data['wavelength']
                )
            for optical_path in data['optical_path']
            for files in optical_path['files']
            for file in files['file']
            ]


def scan_for_metadata_flatten(base_data_dir, extensions=['txt']):
    xls_files = glob.glob(f'{base_data_dir}/**/*.xlsx', recursive=True)
    ret = list()
    for xls in xls_files:
        try:
            dirname = os.path.dirname(xls)
            data = xlsx_to_dict(os.path.dirname(xls), xls)
            flat = metadata_flatten(data)
            for ff in flat:
                for ext in extensions:
                    ff_cpy = ff.copy()
                    filepath = glob.glob(f"{dirname}/**/{ff_cpy['file']}.{ext}", recursive=True)
                    if len(filepath) == 0:
                        continue
                    elif len(filepath) != 1:
                        print(f'len(filepath) = {len(filepath)}, {filepath}')
                        continue
                    ff_cpy['file'] = filepath[0]
                    ff_cpy['file_format'] = ext
                    ret.append(ff_cpy)
        except Exception as e:
            print(f'{e} {xls}')
    return ret


def read_metadata(root_folder,config_input,product):
    if not os.path.exists(product["data"]):
        os.mkdir(product["data"])
    with open(config_input, 'r') as infile:
        config = json.load(infile)
    for entry in config:
        if entry["enabled"]:
            product_metadata = os.path.join(product["data"],"metadata_{}_{}_{}.json".
                format(entry["hsds_provider"],entry["hsds_instrument"],entry["hsds_wavelength"]))

            ##_root = os.path.join(root_folder,entry["folder_input"]) #not sure why it doesn't work and strips the win part to d:\
            folder_input = os.path.join(root_folder,entry["folder_input"])
            file_metadata = os.path.join(root_folder,entry["file_metadata"])
            get(folder_input,file_metadata,product_metadata)





# + tags=["parameters"]
upstream = []
product = None
probe: None
root: None
reference_spectra: None
twinned_spectra: None 
reference_leds: None 
twinned_leds: None 
sample_spectra: None
sample_test: None
# -

import pandas as pd 
import os.path

def read_template(_path, source_tag = "reference", role_tag="spectra"):
    df = pd.read_excel(_path, sheet_name='Files')
    df.columns = ['sample', 'measurement', 'filename', 'optical_path', 'laser_power_percent', 'laser_power_mw', 'integration_time_ms',
              'humidity', 'temperature', 'date', 'time']
    df_meta = pd.read_excel(_path, sheet_name='Front sheet', skiprows=4)
    df_meta.columns = ['optical_path', 'instrument_make', 'instrument_model', 'wavelength','collection_optics','slit_size','grating','pin_hole_size','collection_fibre_diameter','notes','laser_power_max_mw']    
    df_merged = pd.merge(df, df_meta, on='optical_path', how='left')

    # Open the Excel file and read specific cells directly
    with pd.ExcelFile(_path) as xls:
        provider = xls.parse('Front sheet', usecols="B", nrows=1, header=None).iloc[0, 0]
        investigation = xls.parse('Front sheet', usecols="F", nrows=1, header=None).iloc[0, 0]
    df_merged["provider"] = provider
    df_merged["investigation"] = investigation
    df_merged["source"] = source_tag
    df_merged["role"] = role_tag
    return df_merged

#    df_merged["source"] = key
#    df_merged_list.append(df_merged)

df_spe_ref = read_template(os.path.join(root,reference_spectra),"reference","spectra")
df_spe_ref["path"] = os.path.dirname(reference_spectra)
df_spe_twin = read_template(os.path.join(root,twinned_spectra),"twinned","spectra")
df_spe_twin["path"] = os.path.dirname(twinned_spectra)
df_led_ref = read_template(os.path.join(root,reference_leds),"reference","leds")
df_led_ref["path"] = os.path.dirname(reference_leds)
df_led_twin = read_template(os.path.join(root,twinned_leds),"twinned","leds")
df_led_twin["path"] = os.path.dirname(twinned_leds)

df_merged_list = pd.concat([df_spe_ref, df_spe_twin, df_led_ref, df_led_twin], ignore_index=True)
df_merged_list['id'] = df_merged_list.apply(lambda row: f"{row['provider']}_{row['instrument_make']}_{row['instrument_model']}_{row['wavelength']}_{row['collection_optics']}_{row['investigation']}", axis=1)

df_merged_list.loc[df_merged_list['sample'] == sample_test, 'role'] = 'test'

df_merged_list.to_excel(product["data"],index=False)

#columns = ["instrument_make","instrument_model","wavelength","collection_optics","provider","role"]

#unique_combinations = df_merged_list.loc[df_merged_list["role"]=="spectra"][columns].drop_duplicates()
#unique_combinations

# Filter rows with role "reference"
# will calculate power ratio , we need spectra only, not the LEDs
df = df_merged_list.loc[df_merged_list["role"]=='spectra']
#reference spectra
references = df[df_merged_list['source'] == 'reference' ]
# Merge with rows having the same laser_power_percent and role "twinned"
merged = references.merge(df[df['source'] == 'twinned'], on='laser_power_percent', suffixes=('_ref', '_twin'))
# Calculate ratio
merged['laser_power_mw_ratio'] = merged['laser_power_mw_ref'] / merged['laser_power_mw_twin']
merged['integration_time_ms_ratio'] = merged['integration_time_ms_ref'] / merged['integration_time_ms_twin']

merged[["id_ref","id_twin","laser_power_percent","laser_power_mw_ratio","integration_time_ms_ratio"]].drop_duplicates().to_excel(product["ratios"],index=False)


from collections import UserList

from Orange.data import Table
from AnyQt.QtWidgets import QFileDialog
from Orange.widgets import gui
from Orange.widgets.settings import Setting
from Orange.widgets.widget import Output, OWBaseWidget

import pandas as pd
from Orange.data.pandas_compat import table_from_frame
import os.path
from fuzzywuzzy import fuzz
import re
import ramanchada2 as rc2


tags = { '785nm':"wavelength", 
        '532nm':"wavelength", 
        '785':"wavelength", 
        '532':"wavelength", 
        '150mW':"laser_power",
        '327mW':"laser_power",
        '(200mW)':"laser_power",      
        '20%':"laser_power_percent",   
          '150ms' : "acquisition_time",
          '150msx5ac' : "acquisition_time",
          'Probe': "probe", 
          '20x': "probe", 
          '1': "replicate",
          '2': "replicate",
          '3': "replicate",
          '4': "replicate",
         '5': "replicate",
         'position 1': "replicate",
         'txt' : "extension",
         'Neon' : 'sample',
         'Ti' : 'sample',
         'nCal' : 'sample',
         'sCal' : 'sample',
         'Si' : 'sample',
         'S1' : 'sample',
         'S0B' : 'sample',
         'PST' : 'sample',
         'LED' : 'sample',
         'NIR' : 'sample',
         'OP01' : "optical_path"
         }

_lookup = {
    "Original file" : "Original file",
    "laser_wavelength" : "wavelength",
    "model" : "instrument",
    "title" : "title",
    "device" : "instrument",
    "laser_wavelength" : "wavelength",
    "laser_powerlevel" : "laser_power_percent",
    "intigration times(ms)":"acquisition_time",
    "integration times(ms)":"acquisition_time",
    "integ_time" :  "acquisition_time"      
}
def fuzzy_match(vals,tags):
    parsed = {}
    parsed_similarity= {}
    for val in vals:
        similarity_score = None
        match = None
        for tag in tags:
            _tmp = fuzz.ratio(val,tag)
            if (similarity_score is None) or (similarity_score < _tmp):
                similarity_score =_tmp
                match = tags[tag]
        if not match in parsed:
            parsed[match]  = val
            parsed_similarity[match] = similarity_score
        else:
            if parsed_similarity[match] < similarity_score:
                parsed[match]  = val
                parsed_similarity[match] = similarity_score        
    if "wavelength" in parsed:
         parsed["wavelength"] = re.findall(r'\d+', parsed["wavelength"])[0]
    if "extension" in parsed:
        del parsed["extension"]            
    return (parsed,parsed_similarity)

class LoadFileNames(OWBaseWidget):
    name = "Load File Names"
    description = "Load file names"
    icon = "icons/spectra.svg"
    resizing_enabled = False
    want_main_area = False

    filenames = Setting([])

    class Outputs:
        data = Output("File list", Table, default=False, auto_summary=False)

    def __init__(self):
        super().__init__()
        box = gui.widgetBox(self.controlArea, self.name)
        gui.button(box, self, "Load File names", callback=self.load_file)

    def load_file(self):
        self.filenames, filt = QFileDialog.getOpenFileNames(
            caption='Open spectra',
            directory='',
            filter='All files (*)',
            initialFilter='All files (*)',
            )
        _tmp = {}
        for fn in self.filenames:
            #parent_path, filename = os.path.split(fn)
            file_name = os.path.basename(fn)
            file_name, file_extension = os.path.splitext(file_name)
            basename = re.split(r'[-_()]+', file_name)

            parsed,parsed_similarity = fuzzy_match([s for s in basename if s],tags)
            try:
                spe=rc2.spectrum.from_local_file(fn)
                for m in ["Original file","laser_wavelength","model","title","laser_powerlevel","integration times(ms)","intigration times(ms)","integ_time","device","spectrometer"]:
                    try:
                        parsed[_lookup[m]] = spe.meta[m]
                    except:
                        pass
            except Exception as err:
                print(err)
                pass
            
            _tmp[fn] = parsed

        self.Outputs.data.send(table_from_frame(pd.DataFrame.from_dict(_tmp, orient="index")))
        #self.Outputs.data.send(table_from_frame(pd.DataFrame(_tmp,columns=["path","folder","filename"])))



from typing import Union

import ramanchada2 as rc2
from AnyQt.QtWidgets import QGroupBox
from Orange.widgets import gui
from pydantic import validate_call


class AddBaseline:
    @validate_call(config=dict(arbitrary_types_allowed=True))
    def __init__(self, parent, *,
                 n_freq: tuple[str, Union[None, QGroupBox]],
                 amplitude: tuple[str, Union[None, QGroupBox]],
                 intercept: tuple[str, Union[None, QGroupBox]],
                 slope: tuple[str, Union[None, QGroupBox]],
                 quadratic: tuple[str, Union[None, QGroupBox]],
                 ):
        self._parent = parent

        self._n_freq = n_freq[0]
        if n_freq[1]:
            self.n_freq = 15
            gui.spin(n_freq[1], self._parent, 'n_freq', 3, 5000,
                     callback=self.auto_process, label='num freq')

        self._amplitude = amplitude[0]
        if amplitude[1]:
            gui.doubleSpin(amplitude[1], self._parent, self._amplitude, 0, 5000, decimals=5, step=.01,
                           label='amplitude', callback=self.auto_process)

        self._intercept = intercept[0]
        if intercept[1]:
            gui.doubleSpin(intercept[1], self._parent, self._intercept, -20000, 20000, decimals=5, step=1,
                           label='intercept', callback=self.auto_process)

        self._slope = slope[0]
        if slope[1]:
            gui.doubleSpin(slope[1], self._parent, self._slope, -1000, 1000, decimals=5, step=.001,
                           label='slope', callback=self.auto_process)

        self._quadratic = quadratic[0]
        if quadratic[1]:
            gui.doubleSpin(quadratic[1], self._parent, self._quadratic, -100, 100, decimals=7, step=.000001,
                           label='quadratic', callback=self.auto_process)

    @validate_call(config=dict(arbitrary_types_allowed=True))
    def __call__(self, spe: rc2.spectrum.Spectrum) -> rc2.spectrum.Spectrum:
        return spe.add_baseline(n_freq=self.n_freq,
                                amplitude=self.amplitude,
                                pedestal=0,
                                func=lambda x: self.intercept + x*self.slope + x**2*self.quadratic
                                )

    def auto_process(self):
        self._parent.auto_process()

    @property
    def n_freq(self):
        return getattr(self._parent, self._n_freq)

    @n_freq.setter
    def n_freq(self, val):
        setattr(self._parent, self._n_freq, val)

    @property
    def amplitude(self):
        return getattr(self._parent, self._amplitude)

    @amplitude.setter
    def amplitude(self, val):
        setattr(self._parent, self._amplitude, val)

    @property
    def intercept(self):
        return getattr(self._parent, self._intercept)

    @intercept.setter
    def intercept(self, val):
        setattr(self._parent, self._intercept, val)

    @property
    def slope(self):
        return getattr(self._parent, self._slope)

    @slope.setter
    def slope(self, val):
        setattr(self._parent, self._slope, val)

    @property
    def quadratic(self):
        return getattr(self._parent, self._quadratic)

    @quadratic.setter
    def quadratic(self, val):
        setattr(self._parent, self._quadratic, val)



from Orange.widgets import gui
from Orange.widgets.settings import Setting

from ..base_widget import FilterWidget

available_models = ['savgol', 'wiener', 'median','gauss','lowess','boxcar']


class Smoothing(FilterWidget):
    name = "Smooth spectrum"
    description = "Smooth spectrum"
    icon = "icons/spectra.svg"

    method = Setting('savgol')
    should_auto_proc = Setting(True)
    should_auto_plot = Setting(True)

    savgol_window_length = Setting(5)
    savgol_polyorder = Setting(3)
    

    def __init__(self):
        # Initialize the widget
        super().__init__()
        box = gui.widgetBox(self.controlArea, self.name)
        self.method_combo = gui.comboBox(
            box, self, "method",
            label="Select method:",
            items=available_models,sendSelectedValue=True,
            callback=self.auto_process
        )                  
        self.savgolbox = gui.widgetBox(self.controlArea, "savgol")      
        gui.spin(self.savgolbox, self, 'savgol_window_length', 2, 31, callback=self.auto_process,label="window length")
        gui.spin(self.savgolbox, self, 'savgol_polyorder', 1, 7,step=2, callback=self.auto_process,label="polyorder")      

    def process(self):
        self.out_spe = list()
        for spe in self.in_spe:
            kwargs = {}
            if self.method == "savgol":
               kwargs["window_length"] = self.savgol_window_length
               kwargs["polyorder"] = self.savgol_polyorder

            self.out_spe.append(
                spe.smoothing_RC1(method=self.method,**kwargs)
            )
            
        self.send_outputs()

   


# + tags=["parameters"]
upstream = []
product = None
metadata = None
ramandb_api = None
hsds_investigation = None
dry_run = None
keycloak_server_url = None
keycloak_client_id = None
keycloak_realm_name = None
hs_username = None
hs_password = None

# -


from dependency_injector import containers, providers
from dependency_injector.wiring import Provide, inject
from keycloak import KeycloakOpenID
from pynanomapper.clients.authservice import TokenService
from pynanomapper.clients.service_import import ImportService
import traceback




class Container(containers.DeclarativeContainer):
    kcclient = providers.Singleton(
        KeycloakOpenID,
        server_url=keycloak_server_url,
        client_id=keycloak_client_id,
        realm_name=keycloak_realm_name,
        client_secret_key="secret"
    )

    tokenservice = providers.Factory(
        TokenService,
        kcclient = kcclient
    )

    importservice = providers.Factory(
        ImportService,
        tokenservice = tokenservice,
        ramandb_api = ramandb_api,
        hsds_investigation = hsds_investigation,
        dry_run = dry_run
    )


@inject
def main(importservice = Provide[Container.importservice], ):
    importservice.login(hs_username,hs_password)
    try:
        metadata_file = metadata     
        importservice.import2hsds(metadata_file,product["data"])
    except Exception as err:
        traceback.print_exc()
    finally:
        importservice.logout()


print(__name__)
container = Container()
container.init_resources()
container.wire(modules=[__name__])
main()



from Orange.widgets import gui
from Orange.widgets.settings import Setting

from ..base_widget import FilterWidget
from ..processings.add_noise import AddNoise


class AddNoiseOW(FilterWidget):
    name = "Add Noise"
    description = "add noise"
    icon = "icons/spectra.svg"

    noise_scale = Setting(.01)

    def __init__(self):
        super().__init__()
        box = gui.widgetBox(self.controlArea, self.name)
        self.add_noise = AddNoise(self,
                                  noise_scale=('noise_scale', box))

    def process(self):
        self.out_spe = list()
        for spe in self.in_spe:
            self.out_spe.append(
                self.add_noise(spe)
                )
        self.send_outputs()



from AnyQt.QtWidgets import QAbstractItemView
from Orange.widgets import gui
from Orange.widgets.settings import Setting

from ..base_widget import FilterWidget


class Select(FilterWidget):
    name = "Select"
    description = "select spectra"
    icon = "icons/spectra.svg"

    select_inputs_idx = Setting([])

    def __init__(self):
        super().__init__()
        box = gui.widgetBox(self.controlArea, self.name)
        self.select_box = gui.listBox(box, self, 'select_inputs_idx',
                                      selectionMode=QAbstractItemView.MultiSelection, callback=self.auto_process)

    def input_hook(self):
        while self.select_box.takeItem(0):
            pass
        for spe_i, spe in enumerate(self.in_spe):
            self.select_box.addItem(f'{spe_i}: {spe!r}')

    def process(self):
        self.out_spe = list()
        for i in self.select_inputs_idx:
            self.out_spe.append(self.in_spe[i])
        self.send_outputs()



from AnyQt import QtCore
from AnyQt.QtWidgets import (QLineEdit, QTableWidget, QTableWidgetItem,
                             QTabWidget, QVBoxLayout, QWidget)
from Orange.widgets.settings import Setting

from ..base_widget import FilterWidget


class MetadataEditorOW(FilterWidget):
    name = "Metadata editor"
    description = "Visualize and edit metadata of the spectra"
    icon = "icons/spectra.svg"

    noise_scale = Setting(.01)

    def __init__(self):
        super().__init__()
        self.skip_changes = True
        self.tabw = QTabWidget()
        self.mainArea.layout().addWidget(self.tabw)
        self.tmp_spe = []

    def clear_all_tabs(self):
        for tab_i in reversed(range(self.tabw.count())):
            widget = self.tabw.widget(tab_i)
            widget.deleteLater()
            self.tabw.removeTab(tab_i)

    @QtCore.pyqtSlot(QTableWidgetItem)
    def onChanged(self, it):
        if self.skip_changes:
            return
        cur_tab_idx = self.tabw.currentIndex()
        if cur_tab_idx < 0:
            return
        table = self.tables[cur_tab_idx]
        meta = {}
        for row_i in reversed(range(table.rowCount())):
            if table.item(row_i, 0) is None or table.item(row_i, 1) is None:
                if table.item(row_i, 0) is None and table.item(row_i, 1) is None:
                    continue
                else:
                    return
            key = table.item(row_i, 0).text()
            val = table.item(row_i, 1).text()
            if key == '':
                table.removeRow(row_i)
                continue
            if table.item(row_i, 0) is not None and table.item(row_i, 1) is None:
                return

            if key in meta:
                key = key + '_orig'
            meta[key] = val
        print(meta)
        spe = self.tmp_spe[cur_tab_idx] = self.tmp_spe[cur_tab_idx].__copy__()
        spe.meta = meta
        self.spe_to_table(spe, table)
        self.auto_process()

    def spe_to_table(self, spe, table):
        self.skip_changes = True
        for _ in range(table.columnCount()):
            table.removeColumn(0)
        table.setRowCount(0)
        table.setColumnCount(2)
        table.setHorizontalHeaderItem(0, QTableWidgetItem('Metadata keys'))
        table.setHorizontalHeaderItem(1, QTableWidgetItem('Metadata values'))
        for k, v in spe.meta.serialize().items():
            table.insertRow(table.rowCount())
            table.setItem(table.rowCount()-1, 0, QTableWidgetItem(k))
            table.setItem(table.rowCount()-1, 1, QTableWidgetItem(str(v)))
        table.insertRow(table.rowCount())
        table.resizeColumnToContents(0)
        table.resizeColumnToContents(1)
        self.skip_changes = False

    def input_hook(self):
        self.clear_all_tabs()
        self.tmp_spe = []
        self.tables = []
        for spe_i, spe in enumerate(self.in_spe):
            widg = QWidget()
            layout = QVBoxLayout(widg)
            table = QTableWidget()
            table.itemChanged.connect(self.onChanged)
            self.spe_to_table(spe, table)
            le = QLineEdit(None)
            le.setText(repr(spe))
            le.setReadOnly(True)
            layout.addWidget(le)
            layout.addWidget(table)
            self.tabw.addTab(widg, str(spe_i))
            self.tmp_spe.append(spe)
            self.tables.append(table)
        self.auto_process()

    def process(self):
        self.out_spe = self.tmp_spe
        self.send_outputs()



from Orange.widgets.widget import Input, Msg

from ..base_widget import ArithmeticWidget
from ..base_widget.types import RC2Spectra


class Add(ArithmeticWidget):
    name = "Add"
    description = "add spectra"
    icon = "icons/spectra.svg"

    class Inputs:
        addend1 = Input("Addend1 (RC2Spectra)", RC2Spectra, default=True, auto_summary = False)
        addend2 = Input("Addend2 (RC2Spectra)", RC2Spectra, default=False, auto_summary = False)

    class Warning(ArithmeticWidget.Warning):
        operands_dont_broadcast = Msg('Operands do not broadcast')

    def update_inputs_info(self):
        self.Warning.clear()
        len1 = len(self.addend1) if self.addend1 else None
        len2 = len(self.addend2) if self.addend2 else None
        self.info.set_input_summary(f'{len1} RC2Spectra + {len2} RC2Spectra')

    @Inputs.addend1
    def set_addend1(self, spe):
        if spe:
            self.addend1 = spe
        else:
            self.addend1 = None
        self.update_inputs_info()
        self.auto_process()

    @Inputs.addend2
    def set_addend2(self, spe):
        if spe:
            self.addend2 = spe
        else:
            self.addend2 = None
        self.update_inputs_info()
        self.auto_process()

    def __init__(self):
        super().__init__()
        self.addend1 = None
        self.addend2 = None
        self.info.set_input_summary(self.info.NoInput)

    def process(self):
        self.out_spe = list()
        if self.addend1 and self.addend2:
            len1 = len(self.addend1)
            len2 = len(self.addend2)
            if len1 == len2:
                for i in range(len1):
                    self.out_spe.append(self.addend1[i] + self.addend2[i])
            elif len1 == 1:
                for i in range(len2):
                    self.out_spe.append(self.addend1[0] + self.addend2[i])
            elif len2 == 1:
                for i in range(len1):
                    self.out_spe.append(self.addend1[i] + self.addend2[0])
            else:
                self.Warning.operands_dont_broadcast()
        self.send_outputs()



from Orange.widgets import gui
from Orange.widgets.settings import Setting

from ..base_widget import FilterWidget


class Normalize(FilterWidget):
    name = "Normalize"
    description = "Normalize"
    icon = "icons/spectra.svg"

    method = Setting('minmax')

    def __init__(self):
        super().__init__()
        box = gui.widgetBox(self.controlArea, self.name)
        gui.comboBox(box, self, 'method', sendSelectedValue=True,
                     items=['unity', 'min_unity', 'unity_density', 'minmax'],
                     label='Normalization method', callback=self.auto_process)

    def process(self):
        self.out_spe = list()
        for spe in self.in_spe:
            self.out_spe.append(
                spe.normalize(self.method)
                )
        self.send_outputs()



from Orange.widgets.widget import Input

from ..base_widget import FilterWidget, RC2Spectra


class SetXaxis(FilterWidget):
    name = "Set X-axis"
    description = "Set x-axis of the spectra equal to the calibrated spectrum"
    icon = "icons/spectra.svg"

    def __init__(self):
        super().__init__()
        self.in_spe_calib = None

    class Inputs(FilterWidget.Inputs):
        in_spe_calib = Input("RC2Spectra calib", RC2Spectra, auto_summary = False)

    @Inputs.in_spe_calib
    def set_in_spe_calib(self, spe):
        self.should_auto_plot = False
        if spe:
            if len(spe) != 1:
                raise ValueError('calibration takes only a single spectrum')
            self.in_spe_calib = spe[0]
            self.auto_process()
        else:
            self.in_spe_calib = RC2Spectra()
        self.input_hook()

    def process(self):
        if not self.in_spe_calib or not self.in_spe:
            return
        self.out_spe = list()
        for spe in self.in_spe:
            self.out_spe.append(
                spe.set_new_xaxis(self.in_spe_calib.x)
                )
        self.send_outputs()



from Orange.widgets import gui
from Orange.widgets.settings import Setting
from Orange.widgets.widget import Msg

from ..base_widget import FilterWidget


class RS2WL(FilterWidget):
    name = "RS to WL"
    description = "Raman shift to Wavelength"
    icon = "icons/spectra.svg"

    laser_wl = Setting(785)

    def __init__(self):
        super().__init__()
        box = gui.widgetBox(self.controlArea, self.name)
        gui.doubleSpin(box, self, 'laser_wl', 0, 5000, decimals=5, step=1, callback=self.auto_process,
                       label='Laser Wavelength [nm]')

    class Warning(FilterWidget.Warning):
        x_label_not_ramanshift = Msg('Expected spectra with Raman shift on xaxis')

    def process(self):
        self.out_spe = list()
        for inspe in self.in_spe:
            spe = inspe.shift_cm_1_to_abs_nm_filter(laser_wave_length_nm=self.laser_wl)
            if 'xlabel' in spe.meta.root:
                if spe.meta['xlabel'] != 'Raman shift [cm¯¹]':
                    self.Warning.x_label_not_ramanshift()
                meta_dct = spe.meta.model_dump()
                meta_dct['xlabel'] = 'Wavelength [nm]'
                spe.meta = meta_dct
            self.out_spe.append(spe)
        self.send_outputs()



from Orange.data import Table, Domain, StringVariable
from Orange.widgets import gui
from Orange.widgets.settings import Setting
from AnyQt.QtWidgets import QFileDialog
from Orange.widgets.widget import OWWidget, Input, Output, Msg
from Orange.data.pandas_compat import table_from_frame, table_to_frame
import numpy as np
import logging
from itertools import cycle
import pandas as pd
import ploomber
from ploomber.executors import Serial
from ploomber.spec import DAGSpec
from ploomber import DAG
from ..base_widget import BaseWidget
from .ploomber_base import load_env, env2table
import os.path

for handler in logging.root.handlers[:]:
    logging.root.removeHandler(handler)
logging.basicConfig(handlers=[logging.FileHandler("charisma_ploomber.log", mode='w')], level=logging.NOTSET)
logging.root.setLevel(logging.NOTSET)
log = logging.getLogger("charisma")
log.info("log hijack for debugging")

class PloomberImportWidget(BaseWidget):
    # Define the widget's name, category, and outputs
    name = "CHARISMA database import"
    description = "CHARISMA database import"
    icon = "icons/upload.png"
    priority = 10
    # want_main_area = False
    # resizing_enabled = False
    # proportion = Setting(50)
    commitOnChange = Setting(0)
    should_auto_proc = Setting(False)
    investigation =  Setting("TEST") 

    # label = Setting("")
    yaml_file = os.path.join(os.path.dirname(__file__), "ploomber","workflow_import", "pipeline_import.yaml")
    env_file = os.path.join(os.path.dirname(__file__), "ploomber", "workflow_import","env_ploomber.yaml")


    class Inputs:
        # specify the name of the input and the type
        metadata = Input("Metadata", Table)
        data_env = Input("Settings", Table, default=False)           

    class Outputs:
        # if there are two or more outputs, default=True marks the default output
        data = Output("Processed Data", Table, default=True)

    @Inputs.metadata
    def set_metadata(self, metadata):
        if metadata:
            self.metadata = metadata
        else:
            self.metadata = None
        self.update_inputs_info()
        

    @Inputs.data_env
    def set_data_env(self, data_env):
        self.Error.processing_error.clear()
        if data_env:
            self.data_env = data_env
        else:
            self.data_env = None


    def update_inputs_info(self):
        self.Warning.clear()
        len11 = len(self.metadata) if self.metadata else None
    
        self.info.set_input_summary(f' (Reference {len11} metadata)')

 
    # same class can be initiated for Error and Information messages
    class Warning(OWWidget.Warning):
        warning = Msg("My warning!")

    class Error(OWWidget.Error):
        processing_error = Msg("Processing error(s).")
    
    class Information(OWWidget.Warning):
        success = Msg("Workflow successful")

    def __init__(self):
        # Initialize the widget
        super().__init__()

        # Load the environment dictionary from the env.yaml file
        self.env = load_env(self.env_file,"ploomber_import")
        self.env["input_folder"] = ""
        self.investigation = self.env["hsds_investigation"]
        self.provider = self.env["hsds_provider"]
        self.instrument = self.env["hsds_instrument"]
        self.username = self.env["hs_username"]
        self.password = self.env["hs_password"]

        self.set_data_env(env2table(self.env))
        self.dag=None
        box = gui.widgetBox(self.controlArea, self.name)
        #gui.button(box, self, "Load pipeline", callback=self.load_workflow)
        self.investigation_edit = gui.lineEdit(box, self, 'investigation', label='Investigation')
        self.provider_edit = gui.lineEdit(box, self, 'provider', label='Provider')
        self.instrument_edit = gui.lineEdit(box, self, 'instrument', label='Instrument')

        self.instrument_edit.textChanged.connect(lambda text: self.update_field("hsds_instrument", text))
        self.provider_edit.textChanged.connect(lambda text: self.update_field("hsds_provider", text))
        self.investigation_edit.textChanged.connect(lambda text: self.update_field("hsds_investigation", text))

        box1 = gui.widgetBox(box, "CHARISMA DB login")
        self.user_edit = gui.lineEdit(box1, self, 'username', label='Username')
        self.password_edit = gui.lineEdit(box1, self, 'password', label='Password')
        
        self.user_edit.textChanged.connect(lambda text: self.update_field("hs_username", text))
        self.password_edit.textChanged.connect(lambda text: self.update_field("hs_password", text))

    def update_field(self, field_name, text):
        setattr(self, field_name, text)
        self.env[field_name] = text
        self.set_data_env(env2table(self.env))
   

    def workflow_input_file(self):
        return os.path.join( self.env["output_folder"],'metadata.xlsx')
    
    def prepare_input(self):
        metadata_file = self.workflow_input_file()
        df = table_to_frame(self.metadata,include_metas=True)
        df.set_index("index", inplace=True)
        df["enabled"] = True
        df["delete"] = False
        df["hsds_investigation"] = self.env["hsds_investigation"]
        df["hsds_provider"] = self.env["hsds_provider"]
        df["hsds_instrument"] = self.env["hsds_instrument"]        

        with pd.ExcelWriter(metadata_file, engine='openpyxl') as writer:
            df.to_excel(writer,sheet_name="files", index=True)
            pd.DataFrame({"value": [self.env["input_folder"]]}, index=["path"]).to_excel(writer, sheet_name='paths')        
    

    def load_workflow(self):
        if not self.yaml_file or not self.env_file:
            log.info("Please select both YAML and environment files.")
            return
        try:
            log.info(self.yaml_file,self.env_file,self.env)
            self.dag_spec = DAGSpec(data= self.yaml_file, env = self.env)
            self.dag = self.dag_spec.to_dag()

            log.info(self.dag.status())
            #self.plot_spe()
            log.info("Workflow loaded successfully.")
        except Exception as e:
            log.info(e)
            self.Error.processing_error(f"Error: {str(e)}")

    def run_workflow(self):
        if self.dag is None:
            self.load_workflow()
        
        try:
            log.info("prepare input")
            self.prepare_input()

        except Exception as err:
            log.error(err)
            self.Error.processing_error(f"Error while preparing input: {str(err)}")

        try:
            self.dag.build(force=True)
            #tbd read results
            #self.send_outputs()
            self.Information.success()
            #self.statusBar().showMessage("Workflow executed successfully.")
        except Exception as e:
            log.error(e)
            self.Error.processing_error(f"Error: {str(e)}")
            #self.statusBar().showMessage(f"Error: {str(e)}")   

    def process(self):
        self.run_workflow()

    def send_report(self):
        pass

    def plot_spe(self):    
        pass


if __name__ == "__main__":
    from Orange.widgets.utils.widgetpreview import WidgetPreview
    from Orange.data import Table
    import os
    try:
        WidgetPreview(PloomberImportWidget).run()
    except Exception as err:
        print(err)
        WidgetPreview(PloomberImportWidget).run()



import pandas as pd
from Orange.data import Table
from Orange.data.pandas_compat import table_from_frame
from Orange.widgets import gui
from Orange.widgets.settings import Setting
from Orange.widgets.widget import Output
from PyQt5.QtCore import QEventLoop, QThread, pyqtSignal
from ramanchada2.misc.types.fit_peaks_result import FitPeaksResult
from ramanchada2.spectrum.peaks.fit_peaks import available_models

from ..base_widget import FilterWidget


class FitterThread(QThread):
    finished = pyqtSignal(object)

    def __init__(self, in_spe, peak_profile, should_fit, vary_baseline):
        super().__init__()
        self.in_spe = in_spe
        self.peak_profile = peak_profile
        self.should_fit = should_fit
        self.vary_baseline = vary_baseline
        self.should_break = [False]
        self.setPriority(QThread.LowestPriority)

    def stop(self):
        self.should_break[0] = True

    def run(self):
        self.should_break[0] = False
        out_spe = []
        for spe in self.in_spe:
            out_spe.append(
                spe.fit_peaks_filter(profile=self.peak_profile,
                                     no_fit=not self.should_fit,
                                     vary_baseline=self.vary_baseline,
                                     should_break=self.should_break)
            )
        if self.should_break[0]:
            self.finished.emit([])
        else:
            self.finished.emit(out_spe)


class Fit(FilterWidget):
    name = "Fit Peaks"
    description = "Fit Peaks"
    icon = "icons/spectra.svg"

    should_fit = Setting(True)
    vary_baseline = Setting(False)
    peak_profile = Setting(available_models[0])
    plot_individual_peaks = Setting(False)
    should_auto_proc = Setting(False)

    class Outputs(FilterWidget.Outputs):
        peaks_out = Output("Peaks", Table, default=False)

    def __init__(self):
        super().__init__()
        box = gui.widgetBox(self.controlArea, self.name)

        gui.checkBox(box, self, "should_fit", "Perform fit", callback=self.auto_process)
        gui.checkBox(box, self, "plot_individual_peaks", "Plot individual peaks", callback=self.auto_process)
        gui.checkBox(box, self, "vary_baseline", "Vary baseline", callback=self.auto_process)
        gui.comboBox(box, self, 'peak_profile', sendSelectedValue=True, items=available_models,
                     callback=self.auto_process)
        self.stop_processing_btn = gui.button(box, self, "Kill fitting", callback=self.break_fit)

    def break_fit(self):
        self.thread.stop()
        self.is_processed = False

    def on_calculation_finished(self, result):
        self.out_spe = result
        self.process_btn.setDisabled(False)
        self.stop_processing_btn.setDisabled(True)
        dfs = []
        for spe in self.out_spe:
            df = FitPeaksResult.loads(spe.result).to_dataframe_peaks()
            fn = pd.Series([spe.meta['Original file']]*len(df))
            fn.index = df.index
            df['filename'] = fn
            dfs.append(df)
        if self.out_spe:
            self.Outputs.peaks_out.send(table_from_frame(pd.concat(dfs)))
            self.send_outputs()

    def onDeleteWidget(self):
        if self.thread and self.thread.isRunning():
            self.thread.terminate()
            self.thread.wait()

    def perform_threaded_fit(self):
        self.process_btn.setDisabled(True)
        self.stop_processing_btn.setDisabled(False)
        self.thread = FitterThread(in_spe=self.in_spe,
                                   peak_profile=self.peak_profile,
                                   should_fit=self.should_fit,
                                   vary_baseline=self.vary_baseline,
                                   )
        self.thread.finished.connect(self.on_calculation_finished)
        self.event_loop = QEventLoop()
        self.thread.finished.connect(self.event_loop.exit)
        self.thread.start()
        self.event_loop.exec_()

    def process(self):
        self.perform_threaded_fit()

    def custom_plot(self, ax):
        if self.plot_individual_peaks:
            peaks_ax = ax.twinx()
        else:
            peaks_ax = ax
        for spe in self.out_spe:
            FitPeaksResult.loads(spe.result).plot(peaks_ax, individual_peaks=self.plot_individual_peaks)



# + tags=["parameters"]
upstream = ["read_templates"]
product = None
probe: None
root: None
crop_left: None
crop_right: None
# -

import pandas as pd
import os.path
from ramanchada2.spectrum import from_chada,from_local_file, Spectrum
from pathlib import Path
import matplotlib.pyplot as plt
import numpy as np 
from sklearn.linear_model import LinearRegression

Path(product["data"]).mkdir(parents=True, exist_ok=True)

df = pd.read_excel(upstream["read_templates"]["data"])
ratios = pd.read_excel(upstream["read_templates"]["ratios"])

df.head()

ratios.head()

#reference = df.loc(df["source"]=='reference')
#twinned = df.loc(df["source"]=='twinned')
def findref_pair(id_twin,laser_power_percent):
    _tmp = ratios.loc[(ratios["id_twin"]==id_twin) & (ratios["laser_power_percent"]==laser_power_percent)]
    
    if _tmp.empty:
        return (None,None,None)
    else:
        row = _tmp.iloc[0]
        return (row["id_ref"],row["laser_power_mw_ratio"],row["integration_time_ms_ratio"])

def calc_regression(x,y):
    model = LinearRegression().fit(x,y)
    #print("Intercept:", model.intercept_)
    #print("Slope (Coefficient):", model.coef_[0])    
    return  (model.intercept_,model.coef_[0])

def process(df, pair_normalize = False, baseline = False, ycalibration = False, peak =False):
    #leds have no laser_power etc,  we want them as well
    id_groups = df.groupby(['id','sample','laser_power_percent','integration_time_ms','source','role'],dropna=False)

    data4regression = []
    # Iterate over each group
    for (id, sample,laser_power_percent,integration_time_ms,source,role), group in id_groups:
        print(id,sample,laser_power_percent,integration_time_ms,source,role,group.shape)
        spe = None
        # average spectra
        # these are replicates and should have the same grid (but better to check that)
        
        for index, row in group.iterrows():
            _tmp = from_local_file(os.path.join(root,row["path"],row["filename"]))
            spe = _tmp if spe is None else spe + _tmp
        spe /= group.shape[0]
        spe  = spe.trim_axes(method='x-axis',boundaries=(crop_left,max(spe.y) if crop_right is None else crop_right))

        if pair_normalize:
            fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(12, 3))
            idref,LP_ratio, T_ratio = findref_pair(id,laser_power_percent) 
            spe.plot(ax1, label='average')
            if LP_ratio != None:
               spe = spe * LP_ratio
            if T_ratio != None:
                spe = spe * T_ratio
            spe.plot(ax1, label='normalized')
            plt.title(id)

        if baseline:
            kwargs = {"niter" : 40}            
            spe = spe.subtract_baseline_rc1_snip(**kwargs)
            #spe.plot(ax2,label='baseline')
            

        if ycalibration:
            #tbd
            pass

        if peak:
            intensity_val,_position = calc_peak_intensity(spe )
            print(intensity_val,_position)
            laser_power_mw = group["laser_power_mw"].unique()
            data4regression.append((id,laser_power_mw[0],intensity_val))

    return data4regression
  
def calc_peak_intensity(spe,peak=144,prominence=0.01,fit_peak= False,peak_intensity="height"):
    try:
        boundaries=(peak-50, peak+50)
        boundaries=(65,300)
        spe = spe.trim_axes(method='x-axis', boundaries=boundaries)
        candidates = spe.find_peak_multipeak(prominence=prominence)
        fig, ax = plt.subplots(figsize=(6,2))

        _position = None
        if fit_peak:
            fit_res = spe.fit_peak_multimodel(profile='Voigt', candidates=candidates)
            df = fit_res.to_dataframe_peaks()
            df["sorted"] = abs(df["center"] - peak) #closest peak to 144
            df_sorted = df.sort_values(by='sorted')
            index_left = np.searchsorted(spe.x, df_sorted["center"][0] , side='left', sorter=None)
            index_right = np.searchsorted(spe.x, df_sorted["center"][0] , side='right', sorter=None)
            intensity_val = (spe.y[index_right] + spe.y[index_left])/2.0
            _label = "intensity = {:.3f} {} ={:.3f} amplitude={:.3f} center={:.1f}".format(
                intensity_val,peak_intensity,df_sorted.iloc[0][peak_intensity],
                df_sorted.iloc[0]["amplitude"],df_sorted.iloc[0]["center"])
            _position = df_sorted.iloc[0]["center"]
            spe.plot(ax=ax, fmt=':',label=_label)
            fit_res.plot(ax=ax)            
        else:
            _col = "amplitude"
            peak_list = []
            for c in candidates:
                for p in c.peaks:
                    peak_list.append({_col: p.amplitude, 'position': p.position})
            df_sorted = pd.DataFrame(peak_list)
            df_sorted["sorted"] = abs(df_sorted["position"] - peak) #closest peak to 144
            df_sorted = df_sorted.sort_values(by='sorted')
            intensity_val = df_sorted.iloc[0][_col]
            _label = "{}={:.3f} position={:.1f}".format(_col,intensity_val,df_sorted.iloc[0]["position"])  
            _position = df_sorted.iloc[0]["position"]          
            spe.plot(ax=ax, fmt=':',label=_label)
        #return df_sorted[peak_intensity][0]
        return intensity_val, _position
    except Exception as err:
        print(err)
        return None

#reference leds
process(df.loc[(df["source"]=="reference") & (df["role"]=="leds")])
#reference spectra
data4regression_reference = process(df.loc[(df["source"]=="reference") & (df["role"]=="spectra")], pair_normalize= False, baseline= True,ycalibration= True, peak=True)
#twinned leds
process(df.loc[(df["source"]=="twinned") & (df["role"]=="leds")])
#twinned spectra
data4regression = process(df.loc[(df["source"]=="twinned") & (df["role"]=="spectra")] , pair_normalize = True,baseline= True, ycalibration= True, peak = True)


df = pd.DataFrame(data4regression_reference,columns=["id","laser_power_mw","peak_intensity"])
factor_correction = []
for id, group in df.groupby(['id']):

    _intercept,_slope = calc_regression(group[["laser_power_mw"]].values,group[["peak_intensity"]].values)
    factor_correction.append((id[0],_intercept[0],_slope[0]))
df_fc = pd.DataFrame(factor_correction,columns=["id","intercept","slope"])
with pd.ExcelWriter(os.path.join(product["data"],"data4regression_reference.xlsx"), engine='openpyxl', mode='a',if_sheet_exists='replace') as writer:
    df.to_excel(writer, sheet_name='data4regression', index=False)
    df_fc.to_excel(writer, sheet_name='factor_correction', index=False)


df = pd.DataFrame(data4regression,columns=["id","laser_power_mw","peak_intensity"])
factor_correction = []
for id, group in df.groupby(['id']):
    _intercept,_slope = calc_regression(group[["laser_power_mw"]].values,group[["peak_intensity"]].values)
    factor_correction.append((id[0],_intercept[0],_slope[0]))
df_fc = pd.DataFrame(factor_correction,columns=["id","intercept","slope"])    
with pd.ExcelWriter(os.path.join(product["data"],"data4regression_twinned.xlsx"), engine='openpyxl', mode='a',if_sheet_exists='replace') as writer:
    df.to_excel(writer, sheet_name='data4regression', index=False)
    df_fc.to_excel(writer, sheet_name='factor_correction', index=False)


from Orange.widgets import gui
from Orange.widgets.settings import Setting

from ..base_widget import CreatorWidget
from ..processings.add_baseline import AddBaseline
from ..processings.add_noise import AddNoise
from ..processings.gen_spe import GenSpe


class GenNoisySpe(CreatorWidget):
    name = "Gen Noisy Spectra"
    description = "Generate spectra with baseline and noise"
    icon = "icons/spectra.svg"

    # Generate Spectrum
    spe_xmin = Setting(0)
    spe_xmax = Setting(2000)
    spe_nbins = Setting(1500)
    deltas_combo = Setting('PST')
    deltas = Setting('1: 1')

    # Add baseline
    n_freq = Setting(15)
    amplitude = Setting(2)
    intercept = Setting(10)
    slope = Setting(.01)
    quadratic = Setting(-.000005)

    # Add noise
    noise_scale = Setting(.01)

    def __init__(self):
        super().__init__()
        self.n_spectra = 1
        box = gui.widgetBox(self.controlArea, self.name)
        box_params = gui.widgetBox(box, 'Spe Params')
        self.gen_spe = GenSpe(self,
                              spe_xmin=('spe_xmin', box_params),
                              spe_xmax=('spe_xmax', box_params),
                              spe_nbins=('spe_nbins', box_params),
                              deltas=('deltas', box_params),
                              )
        box_baseline = gui.widgetBox(box, 'Baseline')
        self.add_baseline = AddBaseline(self,
                                        n_freq=('n_freq', box_baseline),
                                        amplitude=('amplitude', box_baseline),
                                        intercept=('intercept', box_baseline),
                                        slope=('slope', box_baseline),
                                        quadratic=('quadratic', box_baseline),
                                        )
        box_noise = gui.widgetBox(box, 'Noise')
        self.add_noise = AddNoise(self,
                                  noise_scale=('noise_scale', box_noise))

        gui.spin(box, self, 'n_spectra', 1, 50, label='Number of spectra', callback=self.auto_process)
        self.auto_process()

    def process(self):
        clear_spe = self.gen_spe()
        self.out_spe = [
            self.add_noise(
                self.add_baseline(
                    clear_spe
                    )
            )
            for _ in range(self.n_spectra)
            ]

        self.send_outputs()



import ramanchada2 as rc2
from AnyQt.QtWidgets import QFileDialog
from Orange.widgets import gui
from Orange.widgets.settings import Setting

import os

from ..base_widget import CreatorWidget


class LoadFile(CreatorWidget):
    name = "Load File"
    description = "Load spectrum from file"
    icon = "icons/spectra.svg"

    filenames = Setting([])
    dataset = Setting("/raw")

    def __init__(self):
        super().__init__()
        box = gui.widgetBox(self.controlArea, self.name)
        gui.button(box, self, "Load File", callback=self.load_file)
        self.fileformat = 'Auto'
        
        gui.comboBox(box, self, 'fileformat', sendSelectedValue=True,
                     items=['Auto', 'spc', 'sp', 'spa', '0', '1', '2',
                            'wdf', 'ngs', 'jdx', 'dx',
                            'txt', 'txtr', 'csv', 'prn', 'rruf',
                            'cha'],
                     label='File format',callback=self.update_visibility)
        self.datasetbox = gui.widgetBox(self.controlArea, "CHADA datasets")    
        self.datasetedit = gui.lineEdit(self.datasetbox, self, "dataset", label="Dataset name  (.cha file):", callback=self.load_file)
        self.datasetbox.hide()

    def update_visibility(self):
            # Show or hide boxes based on the selected method
            if self.fileformat == "cha":
                self.datasetbox.show()
            else:
                self.datasetbox.hide()

    def load_file(self):
        filters = ['TXT (*.txt)',
                   'CSV (*.csv)',
                   'CHADA (*.cha)',
                   'All spectra formats (*.txt *.csv *.spc *.wdf)',
                   'All files (*)',
                   ]
        filenames, filt = QFileDialog.getOpenFileNames(
            caption='Open spectra',
            directory='',
            filter=';;'.join(filters),
            initialFilter=filters[-2],
            )
        if filenames:
            self.filenames = filenames
            self.auto_process()

    def process(self):
        self.out_spe = []
        for fname in self.filenames:
            name, extension = os.path.splitext(fname)
            if extension == ".cha":
                spe = rc2.spectrum.from_chada(fname, dataset=self.dataset)
            else:
                spe = rc2.spectrum.from_local_file(fname,
                                                   filetype=(self.fileformat if self.fileformat != 'Auto' else None),
                                                   )
            meta_dct = spe.meta.model_dump()
            meta_dct['xlabel'] = 'Raman shift [cm¯¹]'
            spe.meta = meta_dct
            self.out_spe.append(spe)
        self.send_outputs()



import pickle
from Orange.widgets import widget, gui
from Orange.widgets.widget import Input, Output
from ramanchada2.protocols.calibration import CalibrationModel
import ramanchada2 as rc2
from ..base_widget import BaseWidget , FilterWidget, RC2Spectra


class ApplyCalibrationModel(FilterWidget):
    name = "x-calibration - Apply"
    description = "Apply x-calibration"
    icon = "icons/spectra.svg"
    priority = 10

    class Inputs(FilterWidget.Inputs):
        calibration_model = Input("Calibration Model", CalibrationModel)

    def __init__(self):
        super().__init__()
        self.calibration_model = None

    @Inputs.calibration_model
    def set_calibration_model(self, calibration_model):
        self.calibration_model = calibration_model


    def apply_calibration_x(self, old_spe: rc2.spectrum.Spectrum, spe_units="cm-1"):
        new_spe = old_spe
        model_units = spe_units
        for model in self.calibration_model.components:
            if model.enabled:
                new_spe = model.process(new_spe, model_units, convert_back=False)
                model_units = model.model_units
        return new_spe

    def process(self):
        self.out_spe = list()
        for spe in self.in_spe:
            self.out_spe.append(
                    self.apply_calibration_x(spe)
                )
        self.is_processed = True            
        self.send_outputs()    

    def custom_plot(self, ax):
        if  self.calibration_model:
            self.calibration_model.plot(ax=self.axes[0])            
        self.axes[0].legend()
        if self.in_spe:
            for spe in self.in_spe:
                spe.plot(ax=self.axes[1],label="original")
    

    def plot_create_axes(self):
        self.axes = self.figure.subplots(nrows=2, sharex=False)
        return self.axes[1]        


from Orange.widgets import gui
from Orange.widgets.settings import Setting

from ..base_widget import FilterWidget


class BaselineRemove(FilterWidget):
    name = "Baseline remove"
    description = "Detect and remove baseline"
    icon = "icons/spectra.svg"

    niter = Setting(40)
    method = Setting("SNIP")
    als_smooth = Setting(7)
    als_lam = Setting(1e5)
    als_p = Setting(0.001)
    window_size = Setting(10)
    should_auto_proc = Setting(True)
    should_auto_plot = Setting(True)    
    
    def __init__(self):
        # Initialize the widget
        super().__init__()
        box = gui.widgetBox(self.controlArea, self.name)
        self.method_combo = gui.comboBox(
            box, self, "method",
            label="Select method:",
            items=["SNIP","ALS","Moving minimum"],sendSelectedValue=True,
            callback=self.auto_process
        )  
        #snipbox = gui.widgetBox(self.controlArea, "SNIP")
        self.niterspin = gui.spin(box, self, 'niter', 0, 100, callback=self.auto_process,label="niter")        
        self.alsbox = gui.widgetBox(self.controlArea, "ALS")      
        gui.doubleSpin(self.alsbox, self, 'als_lam', 1e2, 1e6, decimals=2, step=100, callback=self.auto_process,
                       label='lam')    
        gui.doubleSpin(self.alsbox, self, 'als_p', 0, 5, decimals=4, step=.0001, callback=self.auto_process,
                       label='p')      
        gui.doubleSpin(self.alsbox, self, 'als_smooth', 0.1, 25, decimals=2, step=0.5, callback=self.auto_process,
                       label='smooth')           
        self.mmbox = gui.widgetBox(self.controlArea, "Moving minimum")           
        gui.spin(self.mmbox, self, 'window_size', 0, 5000, callback=self.auto_process)
        self.update_visibility()

    def process(self):
        self.out_spe = list()
        for spe in self.in_spe:
            if self.method == "ALS":
                kwargs = {"niter" : self.niter, 
                          "smooth" : self.als_smooth,
                          "lam" : self.als_lam,
                          "p" : self.als_p}
                self.out_spe.append(
                    spe.subtract_baseline_rc1_als(**kwargs)
                    )                
                pass
            elif self.method == "SNIP":
                kwargs = {"niter" : self.niter }
                self.out_spe.append(
                    spe.subtract_baseline_rc1_snip(**kwargs)
                    )
            else:
                self.out_spe.append(spe.moving_minimum(self.window_size))    
                            
        self.send_outputs()

    def update_visibility(self):
            # Show or hide boxes based on the selected method
            if self.method == "SNIP":
                self.niterspin.show()
                self.mmbox.hide()
                self.alsbox.hide()
            elif self.method == "ALS":
                self.niterspin.show()
                self.alsbox.show()
                self.mmbox.hide()
            else:
                self.alsbox.hide()
                self.mmbox.show()
                self.niterspin.hide()

    def auto_process(self):
        self.update_visibility()
        super().auto_process()
            


from Orange.data import Table, Domain, ContinuousVariable
from Orange.widgets import gui
from Orange.widgets.settings import Setting
from Orange.widgets.widget import OWWidget, Input, Output, Msg
from ramanchada2.spectrum import Spectrum
import numpy as np
import logging
from itertools import cycle


log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)

class ProcessSpectraWidget:  # (OWWidget):
    # Define the widget's name, category, and outputs
    name = "Process Spectra"
    description = "Process spectra provided as Orange data table"
    icon = "icons/spectra.svg"
    priority = 10
    # want_main_area = False
    # resizing_enabled = False
    # proportion = Setting(50)
    commitOnChange = Setting(0)
    # label = Setting("")

    class Inputs:
        # specify the name of the input and the type
        data = Input("Data", Table)

    class Outputs:
        # if there are two or more outputs, default=True marks the default output
        data = Output("Processed Data", Table, default=True)
        peaks = Output("Peaks", Table, default=False)
        model = Output("Model", Table, default=False)

    # same class can be initiated for Error and Information messages
    class Warning(OWWidget.Warning):
        warning = Msg("My warning!")

    class Error(OWWidget.Error):
        processing_error = Msg("Processing error(s).")

    def __init__(self):
        # Initialize the widget
        super().__init__()
        self.data = None
        self.optionsBox = gui.widgetBox(self.controlArea, "Options")
        gui.checkBox(
            self.optionsBox, self, "commitOnChange", "Commit data on selection change"
        )
        gui.button(self.optionsBox, self, "Commit", callback=self.commit)
        self.optionsBox.setDisabled(False)

        # logging.warning('warning')
        # logging.error('error')
        # logging.exception('exp')

    @Inputs.data
    def set_data(self, data):
        self.Error.processing_error.clear()
        if data:
            self.data = data
        else:
            self.data = None

    def commit(self):
        processed_data, peaks, peaks_data = self.process_data()
        self.Outputs.data.send(processed_data)
        self.Outputs.peaks.send(peaks)
        self.Outputs.model.send(peaks_data)

    def send_report(self):
        # self.report_plot() includes visualizations in the report
        # self.report_caption(self.label)
        pass

    def find_peaks(self, spe, index, moving_minimum_window=None):
        bgm = spe.bayesian_gaussian_mixture(
            n_samples=20000,
            n_components=20,
            max_iter=1000,
            moving_minimum_window=moving_minimum_window,
            random_state=42,
            # trim_range=trim_range
            )
        bgm_peaks = [[mean[0], np.sqrt(cov[0][0]), weight, i]
                     for mean, cov, weight, i in
                     zip(bgm.means_, bgm.covariances_, bgm.weights_, cycle([index]))]
        bgm_peaks = sorted(bgm_peaks, key=lambda x: x[2], reverse=True)
        n_peaks = (np.round(bgm.weights_, 2) > 0).sum()
        bgm_peaks = bgm_peaks[:n_peaks]
        return bgm, bgm_peaks

    def peaks2table(self, peaks):
        attrs = [ContinuousVariable("mean"), ContinuousVariable("sigma"),
                 ContinuousVariable("weight"), ContinuousVariable("index")]
        log.info(peaks)
        return Table.from_numpy(Domain(attrs), peaks)

    def bgm2spe(self, x, bgm_peaks, threshold=300):
        import scipy.stats as stats
        spey = None
        for peak in bgm_peaks:
            if peak[1] < threshold:
                continue
            gm = stats.norm(peak[0], peak[1])
            tmp = peak[2]*gm.pdf(x)
            if spey is None:
                spey = tmp
            else:
                spey = spey + tmp
        return spey

    def process_data(self):
        x = np.array([float(a.name) for a in self.data.domain.attributes])
        domain = None
        table_processed = self.data
        table_peaks = None
        peaks = []
        try:
            n_rows = self.data.X.shape[0]
            for i in np.arange(0, n_rows):
                spe = Spectrum(x=x, y=self.data[i].x)
                spe1 = spe.trim_axes(method='x-axis', boundaries=(140, np.max(x)))
                log.info(len(spe1.x))
                if domain is None:
                    attrs = [ContinuousVariable.make("%f" % f, number_of_decimals=1) for f in spe1.x]
                    domain = Domain(attrs, class_vars=self.data.domain.class_vars,
                                    metas=self.data.domain.metas)
                    table_processed = Table.from_table(domain, self.data)
                    attrs1 = [ContinuousVariable.make("%f" % f, number_of_decimals=1) for f in spe1.x]
                    domain1 = Domain(attrs1)
                    table_peaks = Table.from_domain(domain1, n_rows=n_rows)
                bgm, bgm_peaks = self.find_peaks(spe1, i, moving_minimum_window=50)
                peaks.extend(bgm_peaks)
                y_bgm = self.bgm2spe(spe1.x, bgm_peaks, threshold=0)

                # log.debug(max(spe1.y))
                # spe1 = spe1.subtract_moving_minimum(50)
                spe1 = spe1.hht_sharpening(movmin=50)
                spe1 = spe1.normalize('unity_area')
                # print(cand)
                inst = table_processed[i]
                log.info("{} {} {} {}".format(len(spe1.x), len(spe1.y), len(y_bgm), len(inst.x)))
                inst.x[:] = spe1.y
                inst1 = table_peaks[i]
                inst1.x[:] = y_bgm
                # inst.x[:]=spe1.y[:]

            return table_processed, self.peaks2table(peaks), table_peaks
        except Exception as err:
            log.exception(err)
            # self.Error(str(err))
            # raise Exception
        return None, None, None


if __name__ == "__main__":
    from Orange.widgets.utils.widgetpreview import WidgetPreview
    from Orange.data import Table
    import os
    try:
        log.info(os.getcwdb())
        path = r'../datasets/PST_WiICV532_Z005OP02_005_300msx10.txt'
        # with open(path, "r") as f:
        #     data = f.read()
        log.warning('start')
        # no reader!
        # table = Table.from_file(path)
        data = Table("../datasets/collagen.csv")

        # Send the data table to the output
        WidgetPreview(ProcessSpectraWidget).run(set_data=Table.from_table_rows(data, [0, 1]))
    except Exception as err:
        print(err)
        WidgetPreview(ProcessSpectraWidget).run()



from Orange.data import Table, Domain, StringVariable
from Orange.widgets import gui
from Orange.widgets.settings import Setting
from AnyQt.QtWidgets import QFileDialog
from Orange.widgets.widget import Msg, Input, Output, OWBaseWidget, OWWidget
from Orange.data.pandas_compat import table_from_frame
import numpy as np
import logging
from itertools import cycle
import pandas as pd
import ploomber
from ploomber.executors import Serial
from ploomber.spec import DAGSpec
from ploomber import DAG
from ..base_widget import BaseWidget




class PloomberExternalWidget(BaseWidget):
    # Define the widget's name, category, and outputs
    name = "Ploomber Workflow Runner"
    description = "Execute Ploomber workflows with YAML and environment files."
    icon = "icons/ploomber.png"
    priority = 10
    # want_main_area = False
    # resizing_enabled = False
    # proportion = Setting(50)
    commitOnChange = Setting(0)
    # label = Setting("")

    class Inputs:
        # specify the name of the input and the type
        data = Input("Data", Table)

    class Outputs:
        # if there are two or more outputs, default=True marks the default output
        data = Output("Processed Data", Table, default=True)

    # same class can be initiated for Error and Information messages
    class Warning(OWWidget.Warning):
        warning = Msg("My warning!")

    class Error(OWWidget.Error):
        processing_error = Msg("Processing error(s).")

    def __init__(self):
        # Initialize the widget
        super().__init__()
        self.data = None
        self.dag=None
        box = gui.widgetBox(self.controlArea, self.name)
        gui.button(box, self, "Select pipeline YAML File", callback=self.load_file_pipeline)
        gui.button(box, self, "Select ENV File", callback=self.load_file_env)
        
        gui.button(box, self, "Load pipeline", callback=self.load_workflow)
        gui.button(box, self, "Run pipeline", callback=self.run_workflow)
        #gui.button(self.optionsBox, self, "Commit", callback=self.commit)
        #self.optionsBox.setDisabled(False)

    def load_file_pipeline(self):
        filenames, filt = QFileDialog.getOpenFileNames(
            caption='Select YAML File',
            directory='',
            filter='YAML Files (*.yaml *.yml);;All Files (*)',
            initialFilter='All files (*)',
            )
        if filenames:
            self.yaml_file = filenames[0]
            #domain = Domain([StringVariable("File Name")])
            #data = Table(domain, [(os.path.basename(filename),) for filename in self.filenames])
            #self.Outputs.data.send(data)
            df = pd.DataFrame(self.yaml_file, columns=["filename"])
            #df = pd.DataFrame({"a" : {"col1" : "val1","col2" :"val2"}})
            self.Outputs.data.send(table_from_frame(df))   

    def load_file_env(self):
        filenames, filt = QFileDialog.getOpenFileNames(
            caption='Select ENV File',
            directory='',
            filter='YAML Files (*.yaml *.yml);;All Files (*)',
            initialFilter='All files (*)',
            )
        if filenames:
            self.env_file = filenames[0]
            df = pd.DataFrame(self.env_file, columns=["filename"])
            #df = pd.DataFrame({"a" : {"col1" : "val1","col2" :"val2"}})
            self.Outputs.data.send(table_from_frame(df))               

    def run_workflow_(self):
        df = pd.DataFrame({"b" : {"xxx" : "val1","yyy" :"val2"}})
        self.Outputs.data.send(table_from_frame(df))

    def load_workflow(self):
        if not self.yaml_file or not self.env_file:
            self.statusBar().showMessage("Please select both YAML and environment files.")
            return
        try:
            self.dag_spec = DAGSpec(data= self.yaml_file, env = self.env_file)
            self.dag = self.dag_spec.to_dag()
            log.info(self.dag.status())
           
            self.statusBar().showMessage("Workflow loaded successfully.")
        except Exception as e:
            log.info(e)
            self.statusBar().showMessage(f"Error: {str(e)}")

    def run_workflow(self):
        if self.dag is None:
            self.statusBar().showMessage("Please load the pipelinefirst,then click Run.")
            return
        try:
            self.dag.build()
           
            self.statusBar().showMessage("Workflow executed successfully.")
        except Exception as e:
            log.info(e)
            self.statusBar().showMessage(f"Error: {str(e)}")            

    @Inputs.data
    def set_data(self, data):
        self.Error.processing_error.clear()
        if data:
            self.data = data
        else:
            self.data = None

    def plot_spe(self):
        pass

    def send_report(self):
        pass


if __name__ == "__main__":
    from Orange.widgets.utils.widgetpreview import WidgetPreview
    from Orange.data import Table
    import os
    try:
        WidgetPreview(PloomberExternalWidget).run()
    except Exception as err:
        print(err)
        WidgetPreview(PloomberExternalWidget).run()



# + tags=["parameters"]
upstream = ["load_spectra"]
product = None
probe: None
result_spectrum = "spectrum_normalized"

# -

import pandas as pd
import matplotlib.pyplot as plt

def score_laserpower(reference,twinned):
    # Merge the two DataFrames on "laser_power_percent"
    merged_df = twinned.merge(reference, on="laser_power_percent", suffixes=('', '_ref'))
    # Calculate the score and assign it to the "score" column
    merged_df['score'] = merged_df['laser_power_ref'] / merged_df['laser_power']
    # Update the original "twinned" DataFrame with the calculated scores
    #display(merged_df)
    return merged_df['score'].values



def normalize_spectra(row):
    return row["spectrum"] * row["score"]

devices_h5file= upstream["load_spectra"]["data"]
devices = pd.read_hdf(devices_h5file, key="devices")
devices.head()

leds = pd.read_hdf(devices_h5file, key="led")
leds.head()

def calculate_twinned_score(devices,reference_condition,twinned_condition):
    for group, group_df in devices.loc[reference_condition].groupby(['device', 'instrument', 'probe']):
        probe = group[2]
        #print("-{}-".format(probe))
        twinned = devices.loc[twinned_condition]
        score = score_laserpower(group_df,twinned)
        devices.loc[twinned_condition,"score"] = score
        #get_slope(group_df,twinned)

#score for the referencespectra
devices.loc[devices["reference"],"score"]  =1.0
#score for twinned spectra
twinned_condition = (~devices["reference"]) & (devices["probe"] == probe)
reference_condition = (devices["reference"]) & (devices["probe"] == probe)
calculate_twinned_score(devices,reference_condition,twinned_condition)

print(devices.columns)
# Assert that the DataFrame contains the expected columns
assert set(["score"]).issubset(devices.columns), "score column is missing"

#normalisation
twinned_condition = (~devices["reference"]) & pd.notna(devices["score"])
devices.loc[twinned_condition, result_spectrum] = devices.loc[twinned_condition].apply(normalize_spectra, axis=1)

devices.loc[reference_condition, result_spectrum] = devices.loc[reference_condition].apply(normalize_spectra, axis=1)

devices.loc[twinned_condition | reference_condition][["device","score","laser_power","laser_power_percent"]]

print(devices.columns)
# Assert that the DataFrame contains the expected columns
assert set([result_spectrum]).issubset(devices.columns), "a processed spectrum column is missing"


devices.to_hdf(product["data"], key='devices', mode='w')

leds.to_hdf(product["data"], key='led', mode='a')

pd.DataFrame({"original" : {"field" : "spectrum"}, "normalized" : {"field"  : result_spectrum}}).T.to_hdf(product["data"], key='processing', mode='a')



from Orange.widgets import gui
from Orange.widgets.settings import Setting
from Orange.data import Table
from Orange.widgets.widget import Msg, Input
import numpy as np

from ..base_widget import BaseWidget

from ramanchada2.spectrum import Spectrum

class Table2RCSpectra(BaseWidget):
    name = "Table to RCSpectra"
    description = "Converts table to set of RC2Spectra"
    icon = "icons/spectra.svg"

    should_auto_proc = Setting(True)
    should_auto_plot = Setting(False)

    class Inputs:
        data = Input("Data", Table, default=True)

    def __init__(self):
        # Initialize the widget
        super().__init__()
        self.data = None
        box = gui.widgetBox(self.controlArea, self.name)
                   
    @Inputs.data
    def set_data(self, data):
        self.data = data
        if self.should_auto_proc:
            self.process()

    def process(self):
        self.clear_messages()
        if self.data is None:
            self.error("No data input.")
            return
          
        self.out_spe = list()

        # Filter attributes with names that parse to numbers
        numeric_attrs = []
        numeric_indices = []

        meta_keys = []        

        for i, attr in enumerate(self.data.domain.attributes):
            try:
                x_value = float(attr.name.replace("- Mean","").strip())
                numeric_attrs.append(x_value)
                numeric_indices.append(i)
            except ValueError:
                meta_keys.append(attr.name)

        # Extract metadata values from both metadata and non-numeric attributes
        metadata_values = []
        for row in self.data:
            metadata = {key: row[key] for key in meta_keys}
            for meta_attr in self.data.domain.metas:
                metadata[meta_attr.name] = row[meta_attr]
            metadata_values.append(metadata)

        x = np.array(numeric_attrs)

        self.out_spe = []
        for row, metadata in zip(self.data, metadata_values):
            y = [row[i] for i in numeric_indices]

            # Remove pairs where y is None
            x_filtered = np.array([x[i] for i in range(len(y)) if y[i] is not None])
            y_filtered = np.array([val for val in y if val is not None])
            self.out_spe.append(Spectrum(x_filtered, y_filtered, metadata=metadata))
        self.send_outputs()




from Orange.widgets import gui
from Orange.widgets.settings import Setting
from Orange.widgets.widget import Msg
from scipy import signal

from ..base_widget import FilterWidget


class Resample_NUDFT(FilterWidget):
    name = "Resample NUDFT"
    description = "Resample Non-Uniform Discrete Fourier Transform"
    icon = "icons/spectra.svg"

    xmin = Setting(0)
    xmax = Setting(4000)
    nbins = Setting(100)
    window_function = Setting('blackmanharris')

    class Warning(FilterWidget.Warning):
        small_boundaries = Msg('Provided spectra exceed x-min/x-max limits')

    def __init__(self):
        super().__init__()
        box = gui.widgetBox(self.controlArea, self.name)
        gui.spin(box, self, 'xmin', -1000, 10000, callback=self.auto_process, label='x-min')
        gui.spin(box, self, 'xmax', -1000, 10000, callback=self.auto_process, label='x-max')
        gui.spin(box, self, 'nbins', 1, 10000, callback=self.auto_process, label='n-bins')
        gui.comboBox(box, self, 'window_function', label='window', sendSelectedValue=True,
                     items=['barthann',
                            'bartlett',
                            'blackman',
                            'blackmanharris',
                            'bohman',
                            'boxcar',
                            'hamming',
                            'hann',
                            'nuttall',
                            'parzen',
                            'triang',
                            ], callback=self.auto_process)

    def process(self):
        self.Warning.clear()
        self.out_spe = list()
        for spe in self.in_spe:
            if self.xmin > spe.x.min() or self.xmax < spe.x.max():
                self.Warning.small_boundaries()
            self.out_spe.append(
                spe.resample_NUDFT_filter(x_range=(self.xmin, self.xmax), xnew_bins=self.nbins,
                                          window=getattr(signal.windows, self.window_function))
                )
        self.send_outputs()

    def custom_plot(self, ax):
        for spe in self.in_spe:
            spe.trim_axes(method="x-axis",boundaries=(self.xmin, self.xmax)).plot(ax=ax,label='original') 
        ax.legend()


from Orange.widgets import gui
from Orange.widgets.settings import Setting

from ..base_widget import FilterWidget


class Crop(FilterWidget):
    name = "Crop spectrum"
    description = "Crop spectrum"
    icon = "icons/spectra.svg"

    left = Setting(100)
    right = Setting(3500)
    should_auto_proc = Setting(True)
    should_auto_plot = Setting(True)

    def __init__(self):
        # Initialize the widget
        super().__init__()
        box = gui.widgetBox(self.controlArea, self.name)
        self.left_spin = gui.doubleSpin(box, self, 'left', 0, 4500 , decimals=2, step=10, callback=self.auto_process,
                       label='Left')      
        self.right_spin = gui.doubleSpin(box, self, 'right', 0, 4500 , decimals=2, step=100, callback=self.auto_process,
                       label='Right')                   

    def process(self):
        self.out_spe = list()
        for spe in self.in_spe:
            self.out_spe.append(
                spe.trim_axes(method='x-axis',boundaries=(self.left,self.right))
            )
            
        self.send_outputs()

    def check_values(self):
        # Ensure that right is always greater than left
        if self.right <= self.left:
            self.right = self.left + 1

        # Update the spin boxes to reflect the changes
        self.right_spin.setValue(self.right)
        self.left_spin.setValue(self.left)

        self.auto_process(self)        


# + tags=["parameters"]
upstream = []
product = None
hsds_investigation = None
hs_endpoint = None
ramandb_api = None
hs_username = None
hs_password = None
keycloak_server_url = None
keycloak_client_id = None
keycloak_realm_name = None

# -

from dependency_injector import containers, providers
from dependency_injector.wiring import Provide, inject
from keycloak import KeycloakOpenID
from pynanomapper.clients.authservice import TokenService, QueryService


class Container(containers.DeclarativeContainer):
    kcclient = providers.Singleton(
        KeycloakOpenID,
        server_url=keycloak_server_url,
        client_id=keycloak_client_id,
        realm_name=keycloak_realm_name,
        client_secret_key="secret"
    )

    tokenservice = providers.Factory(
        TokenService,
        kcclient = kcclient
    )

    queryservice = providers.Factory(
        QueryService,
        tokenservice = tokenservice
    )


@inject
def main(qs = Provide[Container.queryservice]):
    try:
        qs.login(hs_username,hs_password)
        try:
            api_metadata = "{}metadata".format(ramandb_api)
            print(api_metadata)
            response = qs.get(api_metadata, params={"domain" : "/{}/".format(hsds_investigation)})
            print(response.json())
        except Exception as err:
            print(err)
        try:
            response = qs.get(hs_endpoint+"/domains", params={"domain" : "/{}/".format(hsds_investigation)})
            print(response.json())
        except Exception as err:
            print(err)

    except Exception as err:
        print(err)
    finally:
        qs.logout()


print(__name__)
container = Container()
container.init_resources()
container.wire(modules=[__name__])
main()



import ramanchada2 as rc2
from AnyQt.QtWidgets import QAbstractItemView
from Orange.widgets import gui
from Orange.widgets.settings import Setting
from ramanchada2.auxiliary.spectra import datasets2 as data

from ..base_widget import CreatorWidget


class TestSpectra(CreatorWidget):
    name = "Load Test Spectra"
    description = "load test spectra from ramanchada2"
    icon = "icons/spectra.svg"

    filters = Setting({})
    selected_filenames = Setting([])

    def __init__(self):
        super().__init__()
        self.list_boxes = {}
        box = gui.widgetBox(self.controlArea, self.name)
        for k, v in data.get_filters().items():
            var = 'listbox_' + k + '_select_idx'
            setattr(self, var, list())
            gui.label(box, self, k)
            self.list_boxes[k] = gui.listBox(
                box, self, var, selectionMode=QAbstractItemView.MultiSelection,
                callback=self.update_filters)
            for item in sorted(v):
                self.list_boxes[k].addItem(item)

        gui.label(box, self, 'filenames')

        self.listbox_filename_select_idx = []
        self.list_box_filename = gui.listBox(
            box, self, 'listbox_filename_select_idx', selectionMode=QAbstractItemView.MultiSelection,
            callback=self.select_filename)

        filenames = sorted(data.get_filenames(**self.filters))
        for item in filenames:
            self.list_box_filename.addItem(item)
        self.auto_process()

    def update_filters(self):
        self.filters = {}
        for k, lb in self.list_boxes.items():
            self.filters[k] = [item.text() for item in self.list_boxes[k].selectedItems()]
        filenames = sorted(data.get_filenames(**self.filters))
        self.list_box_filename.clear()
        for item in filenames:
            self.list_box_filename.addItem(item)

    def select_filename(self):
        self.selected_filenames = [fn_item.text() for fn_item in self.list_box_filename.selectedItems()]
        self.auto_process()

    def process(self):
        self.out_spe = list()
        for fn in data.prepend_prefix(self.selected_filenames):
            spe = rc2.spectrum.from_local_file(fn)
            meta_dct = spe.meta.model_dump()
            meta_dct['xlabel'] = 'Raman shift [cm¯¹]'
            spe.meta = meta_dct
            self.out_spe.append(spe)
        self.send_outputs()



from Orange.widgets import gui
from Orange.widgets.settings import Setting

from ..base_widget import FilterWidget
from ..processings.add_baseline import AddBaseline


class AddBaselineOW(FilterWidget):
    name = "Add Baseline"
    description = "add baseline"
    icon = "icons/spectra.svg"

    n_freq = Setting(15)
    amplitude = Setting(2)
    intercept = Setting(10)
    slope = Setting(.01)
    quadratic = Setting(-.000005)

    def __init__(self):
        super().__init__()
        box = gui.widgetBox(self.controlArea, self.name)
        self.add_baseline = AddBaseline(self,
                                        n_freq=('n_freq', box),
                                        amplitude=('amplitude', box),
                                        intercept=('intercept', box),
                                        slope=('slope', box),
                                        quadratic=('quadratic', box),
                                        )

    def process(self):
        self.out_spe = list()
        for spe in self.in_spe:
            self.out_spe.append(self.add_baseline(spe))
        self.send_outputs()



# + tags=["parameters"]
upstream = ["twinning_normalize"]
product = None
probe: None
wavelength: None
moving_minimum_window: 10
spectrum_to_correct: None
spectrum_corrected_column: None 
baseline_after_ledcorrection: False
baseline_algorithm: None

# -
import matplotlib.pyplot as plt
import ramanchada2 as rc2
import numpy as np
import pandas as pd


def baseline_spectra(spe, algo="als", **kwargs):
    if algo == "snip":
        del kwargs["window"]
        kwargs["niter"]  = 1000
        return spe.subtract_baseline_rc1_snip(**kwargs)
    elif algo == "als":
        del kwargs["window"]
        kwargs["niter"]  = 1000
        kwargs["p"]  = 0.1
        kwargs["lam"]  = 1e3
        #lam: float = 1e5, p: float = 0.001, niter: PositiveInt = 100,#
        return spe.subtract_baseline_rc1_als(**kwargs)
    elif (algo == "none") or (algo is None):  # movingmin
        return spe     
    else:  # movingmin
        window = kwargs.get("window", 10)
        return spe - spe.moving_minimum(window)
    
def Y_532(x):
    A = 8.30752731e-01
    B = 2.54881472e-07
    x0 = 1.42483359e+3
    Y = A * np.exp(-B * (x - x0)**2)
    return Y

def Y_785(x):
    A0 = 5.90134423e-1
    A = 5.52032185e-1
    B = 5.72123096e-7
    x0 = 2.65628776e+3
    Y = A0 + A * np.exp(-B * (x - x0)**2)
    return Y

def Y(x,wavelength=785):
    if wavelength==785:
        return Y_785(x)
    elif wavelength==532:
        return Y_532(x)
    else:
        raise Exception("not supported wavelength {}".format(wavelength))


match_led={
    "reference" : "reference",
    "twinning": "twinning"
}
match_led


devices_h5file= upstream["twinning_normalize"]["data"]
print(devices_h5file)
devices = pd.read_hdf(devices_h5file, "devices")
devices.head()

led_frame = pd.read_hdf(devices_h5file, "led")
led_frame.head()

processing = pd.read_hdf(devices_h5file, "processing")
processing.head()

devices_h5file= product["data"]
devices.head()




def intensity_normalization(row,spectrum_to_correct):
    try:
        spe = row[spectrum_to_correct]        
        spe_y = spe.y
        spe_y[spe_y < 0] = 0
        _Y = Y(spe.x,wavelength)

        subset=row["device"]
        led_row = led_frame.loc[led_frame["device"]==subset]
        spe_dist =led_row["spe_dist"].values[0]
        #spe_led = led_row["spectrum"]
        #spe_dist = led_row["spe_dist"]
        area = led_row["area"].values[0]

        spe_led_sampled= spe_dist.pdf(spe.x)*area
        spe_corrected = _Y*spe_y/spe_led_sampled
        mask = (spe_led_sampled == 0)
        spe_corrected[mask] = 0
        return rc2.spectrum.Spectrum(spe.x, spe_corrected)
    except Exception as err:
        print(err)
        return None


baseline_column = "{}_baseline".format(spectrum_corrected_column)  if baseline_after_ledcorrection else "{}_baseline".format(spectrum_to_correct) 
twinned_condition = (~devices["reference"]) & (devices["probe"] == probe)
reference_condition = (devices["reference"]) & (devices["probe"] == probe)

kwargs= {"window" : moving_minimum_window}
if baseline_after_ledcorrection:
    devices.loc[twinned_condition, spectrum_corrected_column] = devices.loc[twinned_condition].apply(lambda row: intensity_normalization(row,spectrum_to_correct),axis=1)
    devices.loc[reference_condition, spectrum_corrected_column] = devices.loc[reference_condition].apply(lambda row: intensity_normalization(row,spectrum_to_correct),axis=1)
    devices.loc[twinned_condition, baseline_column] = devices.loc[twinned_condition][spectrum_corrected_column].apply(lambda spe : baseline_spectra(spe,baseline_algorithm,**kwargs)) 
    devices.loc[reference_condition, baseline_column] = devices.loc[reference_condition][spectrum_corrected_column].apply(lambda spe : baseline_spectra(spe,baseline_algorithm,**kwargs)) 
    devices.loc[reference_condition, "baseline_removed"] = "after LED correction"    
    devices.loc[twinned_condition, "baseline_removed"] = "after LED correction"    
    processing.loc["led_corrected"] = {"field" : spectrum_corrected_column}
    processing.loc["baseline"] = {"field" : baseline_column}    
else:
    devices.loc[twinned_condition, baseline_column] = devices.loc[twinned_condition][spectrum_to_correct].apply(lambda spe : baseline_spectra(spe,baseline_algorithm,**kwargs)) 
    devices.loc[reference_condition, baseline_column] = devices.loc[reference_condition][spectrum_to_correct].apply(lambda spe : baseline_spectra(spe,baseline_algorithm,**kwargs))     
    devices.loc[twinned_condition, spectrum_corrected_column] = devices.loc[twinned_condition].apply(lambda row: intensity_normalization(row,baseline_column),axis=1)
    devices.loc[reference_condition, spectrum_corrected_column] = devices.loc[reference_condition].apply(lambda row: intensity_normalization(row,baseline_column),axis=1)
    devices.loc[reference_condition, "baseline_removed"] = "before LED correction"    
    devices.loc[twinned_condition, "baseline_removed"] = "before LED correction" 
    processing.loc["baseline"] = {"field" : baseline_column}        
    processing.loc["led_corrected"] = {"field" : spectrum_corrected_column}
        
devices.columns

devices_h5file

devices.to_hdf(devices_h5file, key='devices', mode='w')

processing.to_hdf(devices_h5file, key='processing', mode='a')


led_frame.to_hdf(devices_h5file, key='led', mode='a')

led_frame.columns

# Assert that the DataFrame contains the expected columns
assert set([spectrum_corrected_column,baseline_column]).issubset(devices.columns), "a processed spectrum column is missing"




import numpy as np
import ramanchada2.misc.constants as rc2const
import ramanchada2.misc.utils as rc2utils
from Orange.widgets import gui
from Orange.widgets.settings import Setting
from Orange.widgets.widget import OWWidget, Input, Msg
from ..base_widget import BaseWidget , RC2Spectra
import logging
import ramanchada2 as rc2

from ramanchada2.protocols.calibration import YCalibrationComponent, YCalibrationCertificate, CertificatesDict



class YAxisCalibrationWidget(BaseWidget):
    name = "Y axis calibration"
    description = "Y-axis (intensity) calibration"
    icon = "icons/spectra.svg"

    selected_wavelength = Setting("785")
    selected_certificate = Setting('NIST785_SRM2241')
    should_auto_proc = Setting(True)
    should_auto_plot = Setting(True)

    def input_hook(self):
        pass


    class Inputs:
        in_spe = Input("Spectra to calibrate", RC2Spectra, default=True)
        srm_spe = Input("SRM spectrum", RC2Spectra, default=False)


    @Inputs.in_spe
    def set_in_spe(self, spe):
        if spe:
            self.in_spe = spe
        else:
            self.in_spe = RC2Spectra()
        self.update_inputs_info()

    @Inputs.srm_spe
    def set_srm_spe(self, spe):
        if spe:
            self.srm_spe = spe
        else:
            self.srm_spe = RC2Spectra()
        self.update_inputs_info()

    # same class can be initiated for Error and Information messages
    class Warning(OWWidget.Warning):
        warning = Msg("Warning")

    class Information(OWWidget.Information):
        success = Msg("Workflow successful")

    class Error(OWWidget.Error):
        processing_error = Msg("Workflow error")

    def update_inputs_info(self):
        self.Warning.clear()
        len11 = len(self.in_spe) if self.in_spe else None
        len12 = len(self.srm_spe) if self.srm_spe else None
       
        self.info.set_input_summary(f' (Input {len11} RC2Spectra + {len12} SRM RC2Spectra)')

    def __init__(self):
        super().__init__()
        self.srm_spe = RC2Spectra()
        self.certificates = CertificatesDict()

        box = gui.widgetBox(self.controlArea, self.name)
        self.wavelength_options = self.certificates.get_laser_wl()
        self.wavelength_options = sorted(map(int, self.wavelength_options))
        self.selected_wavelength = 0
        self.wavelength_combo = gui.comboBox(
            box, self, "selected_wavelength",
            label="Select Wavelength:",
            items=self.wavelength_options,
            callback=self.update_certificate_options
        )

        # Second dropdown for selecting SRM certificate based on wavelength
        # Second dropdown for selecting SRM certificate based on wavelength
        self.certificate_combo = gui.comboBox(
            box, self, "selected_certificate",
            label="Select SRM Certificate:",
            items=[],sendSelectedValue=True,
            callback=self.auto_process
        )
        self.update_certificate_options()                
        
    def update_certificate_options(self):
        # Get the selected wavelength
        self.should_auto_proc = False
        selected_wavelength = self.wavelength_options[self.selected_wavelength]
        certs = self.certificates.get_certificates(wavelength=selected_wavelength)
        self.certificate_combo.clear()
        self.certificate_combo.addItems(certs.keys())
        self.certificate_combo.setCurrentIndex(0)
        #print(self.certificate_combo.currentIndex(),self.certificate_combo.currentText())
        self.should_auto_proc = True

    def get_certificate(self):
        selected_wavelength = self.wavelength_options[self.selected_wavelength]
        return self.certificates.get(selected_wavelength,self.certificate_combo.currentText())

    def ycalibrate(self,spe,ycal):
        return ycal.process(spe)

    def process(self):
        selected_wavelength = self.wavelength_options[self.selected_wavelength]
        cert = self.get_certificate()
        ycal = YCalibrationComponent(selected_wavelength, reference_spe_xcalibrated=self.srm_spe[0],certificate=cert)
        print("Processing Y-axis calibration with SRM: {}".format(cert.id))

        self.out_spe = list()
        for spe in self.in_spe:
            self.out_spe.append(
                    self.ycalibrate(spe,ycal)
                )
        self.is_processed = True            
        self.send_outputs()

    def custom_plot(self, ax):

        if not self.in_spe:
            return
        for spe in self.in_spe:
            spe.plot(ax=self.axes[0])
        for spe in self.srm_spe:
            try:
                spe.plot(ax=self.axes[0],color="magenta",label=spe.meta["Original file"])
            except:
                pass
        cert = self.get_certificate()
        if cert != None:
            cert.plot(ax=self.axes[0].twinx(),color="pink")

    

    def plot_create_axes(self):
        self.axes = self.figure.subplots(nrows=2, sharex=True)
        return self.axes[1]

if __name__ == "__main__":
    from Orange.widgets.utils.widgetpreview import WidgetPreview
    import os
    try:
        WidgetPreview(YAxisCalibrationWidget).run()
    except Exception as err:
        print(err)
        WidgetPreview(YAxisCalibrationWidget).run()



import ramanchada2 as rc2

from ..base_widget import FilterWidget


class HDRMerger(FilterWidget):
    name = "HDR Merger"
    description = rc2.spectrum.hdr_from_multi_exposure.__doc__
    icon = "icons/spectra.svg"

    def process(self):
        if self.in_spe:
            self.out_spe = [rc2.spectrum.hdr_from_multi_exposure(self.in_spe)]
        self.send_outputs()



import ramanchada2 as rc2
from AnyQt.QtWidgets import QFileDialog
from Orange.widgets import gui
from Orange.widgets.settings import Setting

import os

from ..base_widget import FilterWidget


class SaveFile(FilterWidget):
    name = "Save Spectra"
    description = "Save spectra to file"
    icon = "icons/spectra.svg"

    filenames = Setting([])
    dataset = Setting("/oranchada")

    def __init__(self):
        super().__init__()
        box = gui.widgetBox(self.controlArea, self.name)
        gui.lineEdit(box, self, "dataset", label="Dataset name:")
              
        gui.button(box, self, "Save File", callback=self.save_file)


    def save_file(self):
        filters = ['CHADA file (*.cha)',
                   'TXT file (*.txt)',
                   'All files (*)']
        filename, filt = QFileDialog.getSaveFileName(
            self, 
            caption='Save spectra',
            directory='',
            filter=';;'.join(filters),
            initialFilter=filters[0]
        )
        if filename:
            _, extension = os.path.splitext(filename)
            if extension.lower() == ".cha":
                for spe in self.in_spe:
                    spe.write_cha(filename,dataset = self.dataset)
            elif extension.lower() == ".txt":
                for spe in self.in_spe:
                    #spe.write_cha(filename,dataset = "/raw")    
                    pass                

    def process(self):
        self.out_spe = []
        for spe in self.in_spe:
            self.out_spe.append(spe)
        self.send_outputs()



import unittest
from widgets_pro.process_spectra import ProcessSpectraWidget


class TestLoadSpectraWidget(unittest.TestCase):
    def test_functionality(self):
        # Create an instance of the widget
        widget = ProcessSpectraWidget()

        # Set the input data and parameters
        widget.inputs["Input Data"] = "some_data"
        #widget.parameters["Parameter 1"] = value1
        #widget.parameters["Parameter 2"] = value2

        # Invoke the apply function
        widget.apply()

        # Check that the output data is as expected
        self.assertEqual(widget.outputs["Output Data"], "")

if __name__ == "__main__":
    unittest.main()



import matplotlib.pyplot as plt
import pandas as pd
from AnyQt.QtWidgets import QSizePolicy
from matplotlib.backends.backend_qt5agg import \
    FigureCanvasQTAgg as FigureCanvas
from matplotlib.backends.backend_qt5agg import \
    NavigationToolbar2QT as NavigationToolbar
from Orange.data import Table, Domain, ContinuousVariable, Table,StringVariable,DiscreteVariable

from Orange.data.pandas_compat import table_from_frame
from Orange.widgets import gui
from Orange.widgets.settings import Setting
from Orange.widgets.widget import Msg, Output, OWBaseWidget, OWWidget

from .types import RC2Spectra


class BaseWidget(OWBaseWidget, openclass=True):
    resizing_enabled = True
    priority = 10

    should_auto_plot = Setting(False)
    should_auto_proc = Setting(True)
    should_pass_datatable = Setting(False)
    should_plot_legend = Setting(True)

    class Warning(OWWidget.Warning):
        different_x_labels = Msg('When multiple spectra are provided they are expected to have same xlabel')

    def __init__(self):
        super().__init__()
        self.info.set_input_summary(self.info.NoInput)
        self.info.set_output_summary(self.info.NoOutput)
        self.in_spe = RC2Spectra()
        self.figure = None
        self.is_processed = False
        self.controlArea.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Maximum)
        self.optionsBox = gui.widgetBox(self.controlArea, "Main Options")
        self.optionsBox.setDisabled(False)
        self.should_auto_plot_checkbox = gui.checkBox(self.optionsBox, self, "should_auto_plot", "Auto update plot",
                                                      stateWhenDisabled=False, callback=self.force_plot)
        gui.checkBox(self.optionsBox, self, "should_pass_datatable", "Pass datatable",
                     stateWhenDisabled=False, callback=self.process)
        gui.checkBox(self.optionsBox, self, "should_auto_proc", "Auto process",
                     disables=[self.should_auto_plot_checkbox])
        self.process_btn = gui.button(self.optionsBox, self, "Process", callback=self.process)
        gui.button(self.optionsBox, self, "Plot", callback=self.force_plot)
        gui.checkBox(self.optionsBox, self, "should_plot_legend", "Plot legend", callback=self.auto_process)

    class Outputs:
        out_spe = Output("RC2Spectra", RC2Spectra, default=True, auto_summary=False)
        data = Output("Data", Table, default=False)

    def force_plot(self):
        if not self.is_processed:
            self.process()
            self.is_processed = True
        self.plot_spe()

    def plot_create_axes(self):
        return self.figure.add_subplot(111)

    def plot_spe(self):
        if self.figure is None:
            self.figure = plt.figure(tight_layout=True)
            self.canvas = FigureCanvas(self.figure)
            self.mainArea.layout().insertWidget(0, self.canvas)
            self.toolbar = None
        for a in self.figure.axes:
            self.figure.delaxes(a)
        ax = self.plot_create_axes()
        for spe in self.out_spe:
            spe.plot(ax=ax, label=f'id(spe)={id(spe)}')
        self.set_x_title(ax)
        self.custom_plot(ax)
        if self.should_plot_legend:
            ax.legend(fontsize='x-small', ncols=2)
        else:
            ax.legend([])
        if self.toolbar:
            self.toolbar.setParent(None)
            self.toolbar.destroy()
        self.toolbar = NavigationToolbar(self.canvas, self, coordinates=True)
        self.mainArea.layout().insertWidget(0, self.toolbar)
        self.figure.set_tight_layout(True)
        self.canvas.draw()

    def custom_plot(self, ax):
        pass

    def set_x_title(self, ax):
        xlab = ''
        for spe in reversed(self.out_spe):
            if 'xlabel' in spe.meta.root:
                xlab = spe.meta['xlabel']
        if not xlab:
            return
        should_warn = False
        for spe in self.out_spe:
            if 'xlabel' in spe.meta.root:
                if xlab != spe.meta['xlabel']:
                    should_warn = True
            else:
                should_warn = True
        if should_warn:
            self.Warning.different_x_labels()
        ax.set_xlabel(xlab)

    def output_table_old(self):
        # has no metadata
        df = pd.DataFrame([pd.Series(index=spe.x, data=spe.y) for spe in self.out_spe])
        df = df.sort_index(axis='columns')
        df.columns = [f'{i}' for i in df.columns]
        return table_from_frame(df)

    def send_output_table(self):
        if self.should_pass_datatable:
            if self.out_spe:
                try:
                    self.Outputs.data.send(self.output_table_domain())
                except:  # just in case the new implementation breaks
                    self.Outputs.data.send(self.output_table_old())

    def output_table_domain(self):

        # Extract unique x values from all spectra
        unique_x_values = sorted(set(x for spe in self.out_spe for x in spe.x))
        # Extract metadata keys
        all_meta_keys = set()
        for spe in self.out_spe:
            all_meta_keys.update(spe.meta.get_all_keys())

            # Create Orange Table Domain
            domain_variables = [ContinuousVariable(str(x)) for x in unique_x_values]
            meta_variables = [StringVariable(key) for key in all_meta_keys]
            domain = Domain(domain_variables, metas=meta_variables)
            data = []
            for spe in self.out_spe:
                row_data = [None] * len(unique_x_values)  # Initialize row data with None
                for x, y in zip(spe.x, spe.y):
                    column_index = unique_x_values.index(x)
                    row_data[column_index] = y
                meta_values = []
                for key in all_meta_keys:
                    if key in spe.meta.root:
                        meta_values.append(spe.meta[key])
                    else:
                        meta_values.append(None)
                data.append(row_data + meta_values)

            return Table.from_list(domain, data)

    def send_outputs(self):
        if self.out_spe:
            self.Outputs.out_spe.send(self.out_spe)
            self.send_output_table()
            self.info.set_output_summary(f'{len(self.out_spe)} RC2Spectra',
                                         '\n'.join([f'· {repr(i)}' for i in self.out_spe]))
        else:
            self.info.set_output_summary(self.info.NoOutput)

    def process(self):
        self.is_processed = True

    def auto_process(self):
        self.is_processed = False
        if self.should_auto_proc:
            self.process()
            self.is_processed = True
            if self.should_auto_plot:
                self.plot_spe()



from Orange.widgets import gui
from Orange.widgets.settings import Setting

from ..base_widget import FilterWidget


class MovingMinimum(FilterWidget):
    name = "Moving minimum"
    description = "moving minimum"
    icon = "icons/spectra.svg"

    window_size = Setting(10)

    def __init__(self):
        # Initialize the widget
        super().__init__()
        box = gui.widgetBox(self.controlArea, self.name)
        gui.spin(box, self, 'window_size', 0, 5000, callback=self.auto_process)

    def process(self):
        self.out_spe = list()
        for spe in self.in_spe:
            self.out_spe.append(
                spe.moving_minimum(self.window_size)
                )
        self.send_outputs()



from Orange.widgets import gui
from Orange.widgets.settings import Setting

from ..base_widget import CreatorWidget
from ..processings.gen_spe import GenSpe


class GenSpeOW(CreatorWidget):
    name = "Gen Spectra"
    description = "gen spectra"
    icon = "icons/spectra.svg"

    spe_xmin = Setting(0)
    spe_xmax = Setting(2000)
    spe_nbins = Setting(1500)
    deltas_combo = Setting('PST')
    deltas = Setting('1: 1')

    n_spectra = Setting(1)

    def __init__(self):
        super().__init__()
        box = gui.widgetBox(self.controlArea, self.name)
        box_params = gui.widgetBox(box, 'Spe Params')
        self.gen_spe = GenSpe(self,
                              spe_xmin=('spe_xmin', box_params),
                              spe_xmax=('spe_xmax', box_params),
                              spe_nbins=('spe_nbins', box_params),
                              deltas=('deltas', box_params),
                              )

        gui.spin(box, self, 'n_spectra', 1, 50, label='Number of spectra', callback=self.auto_process)
        self.auto_process()

    def process(self):
        self.out_spe = [self.gen_spe()] * self.n_spectra
        self.send_outputs()



from typing import Union

import ramanchada2 as rc2
from AnyQt.QtWidgets import QGroupBox
from Orange.widgets import gui
from pydantic import validate_call


class AddNoise:
    @validate_call(config=dict(arbitrary_types_allowed=True))
    def __init__(self, parent, *,
                 noise_scale: tuple[str, Union[None, QGroupBox]],
                 ):
        self._parent = parent

        self._noise_scale = noise_scale[0]
        if noise_scale[1]:
            gui.doubleSpin(noise_scale[1], self._parent, self._noise_scale, 0, 100, decimals=5, step=.001,
                           callback=self.auto_process)

    def auto_process(self):
        self._parent.auto_process()

    @validate_call(config=dict(arbitrary_types_allowed=True))
    def __call__(self, spe: rc2.spectrum.Spectrum) -> rc2.spectrum.Spectrum:
        return spe.add_poisson_noise(self.noise_scale)

    @property
    def noise_scale(self):
        return getattr(self._parent, self._noise_scale)

    @noise_scale.setter
    def noise_scale(self, val):
        setattr(self._parent, self._noise_scale, val)



# + tags=["parameters"]
upstream = ["twinning_intensity_normalization"]
product = None
probe: None
spectrum_corrected_column: None
baseline_after_ledcorrection: None
fit_peak: None

# -

import pandas as pd
from sklearn.linear_model import LinearRegression
import matplotlib.pyplot as plt
import ramanchada2 as rc2
import numpy as np

peak_intensity="height"
spectra2process = "{}_baseline".format(spectrum_corrected_column)  if baseline_after_ledcorrection else spectrum_corrected_column
print(spectra2process)

def calc_peak_intensity(spe,peak=144,prominence=0.01):
    try:
        boundaries=(peak-50, peak+50)
        spe = spe.trim_axes(method='x-axis', boundaries=boundaries)
        candidates = spe.find_peak_multipeak(prominence=prominence)
        fig, ax = plt.subplots(figsize=(6,2))

        if fit_peak:
            fit_res = spe.fit_peak_multimodel(profile='Voigt', candidates=candidates)
            df = fit_res.to_dataframe_peaks()
            df["sorted"] = abs(df["center"] - peak) #closest peak to 144
            df_sorted = df.sort_values(by='sorted')
            index_left = np.searchsorted(spe.x, df_sorted["center"][0] , side='left', sorter=None)
            index_right = np.searchsorted(spe.x, df_sorted["center"][0] , side='right', sorter=None)
            intensity_val = (spe.y[index_right] + spe.y[index_left])/2.0
            _label = "intensity = {:.3f} {} ={:.3f} amplitude={:.3f} center={:.1f}".format(
                intensity_val,peak_intensity,df_sorted.iloc[0][peak_intensity],
                df_sorted.iloc[0]["amplitude"],df_sorted.iloc[0]["center"])
            spe.plot(ax=ax, fmt=':',label=_label)
            fit_res.plot(ax=ax)            
        else:
            _col = "amplitude"
            peak_list = []
            for c in candidates:
                for p in c.peaks:
                    peak_list.append({_col: p.amplitude, 'position': p.position})
            df_sorted = pd.DataFrame(peak_list)
            df_sorted["sorted"] = abs(df_sorted["position"] - peak) #closest peak to 144
            df_sorted = df_sorted.sort_values(by='sorted')
            intensity_val = df_sorted.iloc[0][_col]
            _label = "{}={:.3f} position={:.1f}".format(_col,intensity_val,df_sorted.iloc[0]["position"])            
            spe.plot(ax=ax, fmt=':',label=_label)
        #return df_sorted[peak_intensity][0]
        return intensity_val
    except Exception as err:
        print(err)
        return None

#slope
def calc_regression(x,y):
    model = LinearRegression().fit(x,y)
    #print("Intercept:", model.intercept_)
    #print("Slope (Coefficient):", model.coef_[0])    
    return  (model.intercept_,model.coef_[0])

devices_h5file= upstream["twinning_intensity_normalization"]["data"]
devices = pd.read_hdf(devices_h5file, "devices")
devices.head()

led_frame = pd.read_hdf(devices_h5file, "led")
led_frame.head()


processing = pd.read_hdf(devices_h5file, "processing")
processing.head()

devices_h5file =product["data"]

#peaks
reference_condition = (devices["reference"]) & (devices["probe"] == probe)
A = devices.loc[reference_condition]
A.head()

devices.loc[reference_condition,peak_intensity] = A[spectra2process].apply(calc_peak_intensity)

twinned_condition = (~devices["reference"]) & (devices["probe"] == probe)
B = devices.loc[twinned_condition]
B.head()

devices.loc[twinned_condition,peak_intensity] = B[spectra2process].apply(calc_peak_intensity)

devices.to_hdf(devices_h5file, key='devices', mode='w')

#regression
A= devices.loc[reference_condition,["device","reference","laser_power",peak_intensity]]
#.dropna()
A

(intercept_A,slope_A) = calc_regression(A[["laser_power"]],A[peak_intensity])

#regression
B= devices.loc[twinned_condition,["device","reference","laser_power",peak_intensity]]
#.dropna()
B

(intercept_B,slope_B) = calc_regression(B[["laser_power"]],B[peak_intensity])


#A = devices.loc[reference_condition]
#B = devices.loc[twinned_condition]

plt.plot(A["laser_power"],A[peak_intensity],'o',label=A["device"].unique())
plt.plot(B["laser_power"],B[peak_intensity],'+',label=B["device"].unique())
plt.legend()

#Factor correction (FC) is obtained by dividing the slope of the reference equipment (spectrometer A) 
# by the slope of the equipment to be twinned (spectrometer B).

factor_correction = slope_A/slope_B
print(slope_A,slope_B,factor_correction)



def harmonization(row):
    try:
        spe = row[spectra2process]       
        return rc2.spectrum.Spectrum(spe.x, spe.y *factor_correction)
    except Exception as err:
        print(err)
        return None
#only twinned is multiplied
devices.loc[twinned_condition,"spectrum_harmonized"] = devices.loc[twinned_condition].apply(harmonization,axis=1)
processing.loc["harmonized"] = {"field" : "spectrum_harmonized"}    
devices.to_hdf(devices_h5file, key='devices', mode='w')


def spe_area(spe):
    try:
        sc = spe.trim_axes(method='x-axis', boundaries=(100, 1750))  
        return np.sum(sc.y * np.diff(sc.x_bin_boundaries))
    except Exception as err:
        print(err)
        return None

devices.loc[reference_condition,"area"] = devices.loc[reference_condition][spectra2process].apply(spe_area)
devices.loc[reference_condition][["reference","device","laser_power","area"]]

devices.loc[twinned_condition,"area"] = devices.loc[twinned_condition][spectra2process].apply(spe_area)   
devices.loc[twinned_condition,"area_harmonized"] = devices.loc[twinned_condition]["spectrum_harmonized"].apply(spe_area)   
devices.loc[twinned_condition][["reference","device","laser_power","area","area_harmonized"]]

devices.to_hdf(devices_h5file, key='devices', mode='w')

processing.to_hdf(devices_h5file, key='processing', mode='a')

led_frame.to_hdf(devices_h5file, key='led', mode='a')

pd.DataFrame({"twinned" : {"slope"  : slope_B, "intercept" : intercept_B},
              "reference" : {"slope"  : slope_A, "intercept" : intercept_A}
              }).T.to_hdf(devices_h5file, key='regression', mode='a')

pd.DataFrame({"result" : {"factor_correction" :factor_correction}
              }).to_hdf(devices_h5file, key='factor_correction', mode='a')

import matplotlib.pyplot as plt
def plot_spectra(row,boundaries=(100, 1750)):
    try:
        sc =row[spectra2process]
        if boundaries:
            sc = sc.trim_axes(method='x-axis', boundaries=boundaries)        
        sc.plot(ax=axes[0],label="{}%".format(row["laser_power_percent"]))
    except:
        pass
    try:
        sc =row["spectrum_harmonized"]
        if boundaries:
            sc = sc.trim_axes(method='x-axis', boundaries=boundaries)
        sc.plot(ax=axes[1],label="{}%".format(row["laser_power_percent"]))
    except:
        pass    
    axes[0].set_title("{} {}".format(row["device"],row["probe"]))

fig, axes = plt.subplots(2, 1, figsize=(15, 6))      
axes[1].set_title("harmonized")
devices.loc[twinned_condition].apply(plot_spectra, axis=1)

fig, axes = plt.subplots(2, 1, figsize=(15, 6))      
axes[1].set_title("harmonized")
devices.loc[reference_condition].apply(plot_spectra, axis=1)
 


# + tags=["parameters"]
import uuid
import json
import h5pyd
from datetime import date
upstream = []
product = None
hs_endpoint = None
hsds_investigation = None
hs_username = None
hs_password = None
keycloak_server_url = None
keycloak_client_id = None
keycloak_realm_name = None

# -

from dependency_injector import containers, providers
from dependency_injector.wiring import Provide, inject
from keycloak import KeycloakOpenID
from pynanomapper.clients.authservice import TokenService
from pynanomapper.clients.service_charisma import H5Service


class Container(containers.DeclarativeContainer):
    kcclient = providers.Singleton(
        KeycloakOpenID,
        server_url=keycloak_server_url,
        client_id=keycloak_client_id,
        realm_name=keycloak_realm_name,
        client_secret_key="secret"
    )
    tokenservice = providers.Factory(
        TokenService,
        kcclient = kcclient
    )
    h5service = providers.Factory(
        H5Service,
        tokenservice = tokenservice
    )


@inject
def main(ts = Provide[Container.h5service]):
    ts.login(hs_admin_username,hs_admin_password)
    try:
        domain = "/{}/".format(hsds_investigation)
        _folder = ts.create_domain(domain)
        print(_folder)

    except Exception as err:
        print(err)
    finally:
        ts.logout()



print(__name__)
container = Container()
container.init_resources()
container.wire(modules=[__name__])
main()



import numpy as np
import ramanchada2.misc.constants as rc2const
import ramanchada2.misc.utils as rc2utils
from Orange.widgets import gui
from Orange.widgets.settings import Setting

from ..base_widget import FilterWidget


class XAxisFineCalibration(FilterWidget):
    name = "Xaxis fine calibration"
    description = "X-axis fine calibration"
    icon = "icons/spectra.svg"

    predefined_deltas = Setting('Ne WL D3.3')
    deltas = Setting('')
    poly_order = Setting('poly3')
    should_fit = Setting(False)
    prominence = Setting(3.)
    wlen = Setting(200)
    min_peak_width = Setting(2)
    should_auto_proc = Setting(False)
    n_iters = Setting(1)

    def __init__(self):
        super().__init__()
        box = gui.widgetBox(self.controlArea, self.name)

        gui.doubleSpin(box, self, 'prominence', 0, 50000, decimals=2, step=1, callback=self.auto_process,
                       label='Prominence [×σ]')
        gui.spin(box, self, 'wlen', 0, 50000, step=1, callback=self.auto_process, label='wlen')
        gui.spin(box, self, 'min_peak_width', 1, 5000, callback=self.auto_process, label='Min peak width')

        gui.checkBox(box, self, "should_fit", "Should fit", callback=self.auto_process)

        gui.comboBox(box, self, 'poly_order', sendSelectedValue=True,
                     items=['poly0', 'poly1', 'poly2', 'poly3', 'poly4', 'RBF thin-plate-spline'],
                     callback=self.auto_process
                     )

        self.combo = gui.comboBox(box, self, 'predefined_deltas', sendSelectedValue=True,
                                  items=[
                                      'Ne RS 532nm', 'Ne RS 633nm', 'Ne RS 785nm',
                                      'Ne WL 532nm', 'Ne WL 633nm', 'Ne WL 785nm',
                                      'Ne WL D3.3', 'PST', 'User defined'], label='Deltas',
                                  callback=self.deltas_combo_callback)
        self.deltas_edit = gui.lineEdit(box, self, 'deltas', callback=self.auto_process)

        gui.spin(box, self, 'n_iters', 0, 20, label='N iters', callback=self.auto_process)

        self.deltas_combo_callback()

    def deltas_combo_callback(self):
        self.deltas_edit.setReadOnly(True)
        if self.predefined_deltas == 'Ne RS 532nm':
            self.deltas = ', '.join([f'{k}: {v}' for k, v in rc2const.neon_rs_532_nist_dict.items()])
        elif self.predefined_deltas == 'Ne RS 633nm':
            self.deltas = ', '.join([f'{k}: {v}' for k, v in rc2const.neon_rs_532_nist_dict.items()])
        elif self.predefined_deltas == 'Ne RS 785nm':
            self.deltas = ', '.join([f'{k}: {v}' for k, v in rc2const.neon_rs_785_nist_dict.items()])
        elif self.predefined_deltas == 'Ne WL 532nm':
            self.deltas = ', '.join([f'{k}: {v}' for k, v in rc2const.neon_wl_532_nist_dict.items()])
        elif self.predefined_deltas == 'Ne WL 633nm':
            self.deltas = ', '.join([f'{k}: {v}' for k, v in rc2const.neon_wl_532_nist_dict.items()])
        elif self.predefined_deltas == 'Ne WL 785nm':
            self.deltas = ', '.join([f'{k}: {v}' for k, v in rc2const.neon_wl_785_nist_dict.items()])
        elif self.predefined_deltas == 'Ne WL D3.3':
            self.deltas = ', '.join([f'{k}: {v}' for k, v in rc2const.neon_wl_D3_3_dict.items()])
        elif self.predefined_deltas == 'PST':
            self.deltas = ', '.join([f'{k}: {v}' for k, v in rc2const.PST_RS_dict.items()])
        elif self.predefined_deltas == 'User defined':
            self.deltas_edit.setReadOnly(False)
        self.auto_process()

    def process(self):
        self.deltas_dict = dict([[float(j) for j in i.split(':')] for i in self.deltas.replace(' ', '').split(',')])
        if self.poly_order == 'poly0':
            poly_order_num = 0
        elif self.poly_order == 'poly1':
            poly_order_num = 1
        elif self.poly_order == 'poly2':
            poly_order_num = 2
        elif self.poly_order == 'poly3':
            poly_order_num = 3
        elif self.poly_order == 'poly4':
            poly_order_num = 4

        self.out_spe = list()
        if len(self.in_spe) > 1:
            raise ValueError(f'Expects a single spectrum input. {len(self.in_spe)} found')
        for spe in self.in_spe:
            for iter in range(self.n_iters):
                if self.poly_order == 'RBF thin-plate-spline':
                    spe = spe.xcal_fine_RBF(ref=self.deltas_dict, should_fit=self.should_fit,
                                            find_peaks_kw=dict(prominence=spe.y_noise*self.prominence,
                                                               wlen=self.wlen,
                                                               width=self.min_peak_width,
                                                               )
                                            )
                else:
                    spe = spe.xcal_fine(ref=self.deltas_dict, poly_order=poly_order_num, should_fit=self.should_fit,
                                        find_peaks_kw=dict(prominence=spe.y_noise*self.prominence,
                                                           wlen=self.wlen,
                                                           width=self.min_peak_width,
                                                           )
                                        )
            self.out_spe.append(spe)
        self.send_outputs()

    def custom_plot(self, ax):
        if not self.in_spe:
            return
        self.in_spe[0].plot(ax=self.axes[0])
        self.axes[0].twinx().stem(list(self.deltas_dict.keys()), list(self.deltas_dict.values()),
                                  basefmt='', linefmt='r:', label='reference')
        ax1_twin = self.axes[1].twinx()
        ax1_twin.stem(list(self.deltas_dict.keys()), list(self.deltas_dict.values()),
                      basefmt='', linefmt='r:', label='reference')
        x, x_exp, y, yerr = self.diff(self.out_spe[0], self.deltas_dict)
        ax1_twin.plot([x, x_exp], [0, 0], 'r', lw=4)
        self.axes[2].errorbar(x, y, yerr, fmt='.:', label='difference')
        for a in self.axes:
            a.grid()
            a.legend()
        y_ub = y + yerr
        y_lb = y - yerr
        y_ub = np.mean(y_ub) + np.std(y_ub)*3
        y_lb = np.mean(y_lb) - np.std(y_lb)*3
        self.axes[2].set_ylim(y_lb, y_ub)
        self.axes[0].set_title('Before calibration')
        self.axes[1].set_title('After calibration')
        self.axes[2].set_title('Difference')
        self.axes[2].set_xlabel(ax.get_xlabel())
        ax.set_xlabel('')

    def plot_create_axes(self):
        self.axes = self.figure.subplots(nrows=3, sharex=True)
        return self.axes[1]

    def diff(self, spe, ref_pos):
        if isinstance(ref_pos, dict):
            ref_pos = list(ref_pos.keys())
        if isinstance(ref_pos, list):
            ref_pos = np.array(ref_pos)
        ss = spe.subtract_moving_minimum(100)
        kw = dict(sharpening=None, hht_chain=[100])
        cand = ss.find_peak_multipeak(**kw)

        if self.should_fit:
            kw = dict(profile='Gaussian')
            fit_res = spe.fit_peak_multimodel(candidates=cand, **kw)
            spe_pos, spe_pos_err = fit_res.centers_err.T
        else:
            spe_pos = np.array(list(cand.get_pos_ampl_dict().keys()))
            spe_pos_err = np.zeros_like(spe_pos)

        spe_pos_match_idx, ref_pos_match_idx = rc2utils.find_closest_pairs_idx(spe_pos, ref_pos)
        spe_pos_match = spe_pos[spe_pos_match_idx]
        ref_pos_match = ref_pos[ref_pos_match_idx]
        return ref_pos_match, spe_pos_match, (spe_pos_match-ref_pos_match), spe_pos_err[spe_pos_match_idx]



# + tags=["parameters"]
upstream = []
product = None
key_spectra = None
key_leds= None
input_h5 = None

# -

import os
import ramanchada2 as rc2
import re
import pandas as pd
import numpy as np

input_spe = pd.read_hdf(input_h5, key=key_spectra)
input_spe["reference"].unique()

input_leds = pd.read_hdf(input_h5, key=key_leds)
input_leds["reference"].unique()

from fuzzywuzzy import fuzz

tags = { '785nm':"wavelength", 
        '532nm':"wavelength", 
        '785':"wavelength", 
        '532':"wavelength", 
        '150mW':"laser_power",
        '327mW':"laser_power",
        '(200mW)':"laser_power",        
          '150ms' : "acquisition_time",
          '150msx5ac' : "acquisition_time",
          'Probe': "probe", 
          '1': "replicate",
          '2': "replicate",
          '3': "replicate",
          '4': "replicate",
         '5': "replicate",
         'txt' : "extension"}

_lookup = {
    "Original file" : "Original file",
    "laser_wavelength" : "wavelength",
    "model" : "instrument",
    "title" : "title",
    "laser_wavelength" : "wavelength",
    "laser_powerlevel" : "laser_power_percent",
    "intigration times(ms)":"acquisition_time"    
}
def fuzzy_match(vals,tags):
    parsed = {}
    parsed_similarity= {}
    for val in vals:
        similarity_score = None
        match = None
        for tag in tags:
            _tmp = fuzz.ratio(val,tag)
            if (similarity_score is None) or (similarity_score < _tmp):
                similarity_score =_tmp
                match = tags[tag]
        if not match in parsed:
            parsed[match]  = val
            parsed_similarity[match] = similarity_score
        else:
            if parsed_similarity[match] < similarity_score:
                parsed[match]  = val
                parsed_similarity[match] = similarity_score        
    if "wavelength" in parsed:
         parsed["wavelength"] = re.findall(r'\d+', parsed["wavelength"])[0]
    if "extension" in parsed:
        del parsed["extension"]            
    return (parsed,parsed_similarity)


def spe_area(spe: rc2.spectrum.Spectrum):
    return np.sum(spe.y * np.diff(spe.x_bin_boundaries))



def load_spectra(input_spe , leds = False):
    spectra={}
    for index,row in input_spe.iterrows():
        spe=row["spectrum"]
        subset = "reference" if row["reference"] else "twinning"
        result = re.split(r'[_().]+', spe.meta["Original file"])
        parsed,parsed_similarity = fuzzy_match([s for s in result if s],tags)
        #print(result)
        for m in ["Original file","laser_wavelength","model","title","laser_powerlevel","intigration times(ms)"]:
            parsed[_lookup[m]] = spe.meta[m]
        #print("parsed",parsed)
        #print("similarity",parsed_similarity)

        laser_power = parsed["laser_power"] if "laser_power" in parsed else None
        laser_power_percent = parsed["laser_power_percent"] if "laser_power_percent" in parsed else None
        acquisition_time = parsed["acquisition_time"] if "acquisition_time" in parsed else None

        if leds:
            tag=subset
        else:
            tag = "{}:{}:{}:{}:{}".format(subset,
                                    parsed["instrument"],parsed["probe"],laser_power,acquisition_time)
        if not tag in spectra:
            spectra[tag] = {"laser_power" : laser_power, "laser_power_percent" :   laser_power_percent,  
                                    "acquisition_time" :  acquisition_time, "reference" : row["reference"],
                                    "spectrum" : spe , "count" : 1 , "probe" : parsed["probe"],"device":subset,
                                    "instrument" :  parsed["instrument"],"title" :  parsed["title"]
                                    ,"wavelength" :  parsed["wavelength"],
                                    "original_file"  : [parsed["Original file"]]
                                    }
        else:
            spectra[tag]["spectrum"] = spectra[tag]["spectrum"] + spe
            spectra[tag]["count"] = spectra[tag]["count"] + 1
            print(spectra[tag])
            spectra[tag]["original_file"].append(parsed["Original file"])
            
        if leds:
            spe_led_y = spe.y
            spe_led_y[spe_led_y < 0] = 0
            spe.y = spe_led_y
            spectra[tag]["spe_dist"] = spe.spe_distribution()
            spectra[tag]["area"] =   spe_area(spe)                   
        #average replicates

    for tag in spectra:
        spe = spectra[tag]["spectrum"]
        spe /= spectra[tag]["count"]
        
    devices = pd.DataFrame(spectra).T
    devices['laser_power'] = devices['laser_power'].str.extract('(\d+)')
    devices['laser_power'] = pd.to_numeric(devices['laser_power'])
    return devices

devices = None
devices =  load_spectra(input_spe)   
devices.to_hdf(product["data"], key='devices', mode='w')

leds = None
leds =  load_spectra(input_leds,leds=True)   
leds.to_hdf(product["data"], key='led', mode='a')

try:
    devices.to_csv(product["meta_spectra"])
    leds.to_csv(product["meta_leds"])
except:
    pass